/*
/
/ sple_internals.c - SpatiaLite - spatial extension for Sqlite - miscellaneous functions
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2007  Alessandro Furieri
/
/ This library is free software; you can redistribute it and/or
/ modify it under the terms of the GNU Lesser General Public
/ License as published by the Free Software Foundation; either
/ version 2.1 of the License, or (at your option) any later version.
/
/ This library is distributed in the hope that it will be useful,
/ but WITHOUT ANY WARRANTY; without even the implied warranty of
/ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/ Lesser General Public License for more details.
/
/ You should have received a copy of the GNU Lesser General Public
/ License along with this library; if not, write to the Free Software
/ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA/
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include "sqlite3ext.h"
#include "spatialite.h"

SPATIALITE_DECLARE int
sple_dimension (sple_geometry_ptr geom)
{
/* determinates the Dimension for this geometry */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    if (n_points == 0 && n_linestrings == 0 && n_polygons == 0)
	return -1;
    if (n_points > 0 && n_linestrings == 0 && n_polygons == 0)
	return 0;
    if (n_linestrings > 0 && n_polygons == 0)
	return 1;
    return 2;
}

SPATIALITE_DECLARE void
sple_geometry_type (sple_geometry_ptr geom, char **result)
{
/* determinates the Class for this geometry */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    *result = NULL;
    if (n_points == 0 && n_linestrings == 0 && n_polygons == 0)
	return;
    *result = malloc (128);
    if (n_points == 1 && n_linestrings == 0 && n_polygons == 0)
      {
	  strcpy (*result, "POINT");
	  return;
      }
    if (n_points > 0 && n_linestrings == 0 && n_polygons == 0)
      {
	  strcpy (*result, "MULTIPOINT");
	  return;
      }
    if (n_points == 0 && n_linestrings == 1 && n_polygons == 0)
      {
	  strcpy (*result, "LINESTRING");
	  return;
      }
    if (n_points == 0 && n_linestrings > 0 && n_polygons == 0)
      {
	  strcpy (*result, "MULTILINESTRING");
	  return;
      }
    if (n_points == 0 && n_linestrings == 0 && n_polygons == 1)
      {
	  strcpy (*result, "POLYGON");
	  return;
      }
    if (n_points == 0 && n_linestrings == 0 && n_polygons > 0)
      {
	  strcpy (*result, "MULTIPOLYGON");
	  return;
      }
    strcpy (*result, "GEOMETRYCOLLECTION");
}

SPATIALITE_DECLARE void
sple_envelope (sple_geometry_ptr geom, unsigned char **result, int *size)
{
/* computes the MBR for this geometry */
    sple_geometry_ptr mbr;
    sple_polygon_ptr polyg;
    sple_linestring_ptr ring;
    sple_mbr_geometry (geom);
    *result = NULL;
    *size = 0;
    if (geom->minx == DBL_MAX && geom->miny == DBL_MAX && geom->maxx == DBL_MIN
	&& geom->maxy == DBL_MAX)
	return;
/* builds a 5 vertices polygon [MBR] */
    polyg = sple_alloc_polygon (5, 0);
    ring = polyg->external;
    sple_set_point (ring, 0, geom->minx, geom->miny);	/* vertex # 1 */
    sple_set_point (ring, 1, geom->maxx, geom->miny);	/* vertex # 2 */
    sple_set_point (ring, 2, geom->maxx, geom->maxy);	/* vertex # 3 */
    sple_set_point (ring, 3, geom->minx, geom->maxy);	/* vertex # 4 */
    sple_set_point (ring, 4, geom->minx, geom->miny);	/* vertex # 5 [same as vertex # 1 to close the polygon] */
/* creates a new geometry containing the polygon */
    mbr = sple_alloc_geometry ();
    mbr->entities = 1;
    mbr->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(mbr->geometries) =
	sple_alloc_generic_geometry (SPLE_POLYGON, (void *) polyg);
    sple_blob_encode (mbr, result, size);	/* generating BLOB internal representation */
    sple_free_geometry (mbr);
}

SPATIALITE_DECLARE int
sple_x (sple_geometry_ptr geom, double *coord)
{
/* returns the X coordinate from a POINT geometry */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point = NULL;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    if (n_linestrings != 0 || n_polygons != 0)
	return 0;
    if (n_points != 1)
	return 0;
    *coord = point->x;
    return 1;
}

SPATIALITE_DECLARE int
sple_y (sple_geometry_ptr geom, double *coord)
{
/* returns the Y coordinate from a POINT geometry */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point = NULL;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    if (n_linestrings != 0 || n_polygons != 0)
	return 0;
    if (n_points != 1)
	return 0;
    *coord = point->y;
    return 1;
}

SPATIALITE_DECLARE int
sple_num_points (sple_geometry_ptr geom, int *points)
{
/* returns the number of vertices from a LINESTRING geometry */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_linestring_ptr line = NULL;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    if (n_points != 0 || n_polygons != 0)
	return 0;
    if (n_linestrings != 1)
	return 0;
    *points = line->points;
    return 1;
}

SPATIALITE_DECLARE void
sple_point_n (sple_geometry_ptr geom, int vertex, unsigned char **result,
	      int *size)
{
/* extracts the first vertex from a LINESTRING geometry */
    int ie;
    int index;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_geometry_ptr new_geom;
    sple_generic_geometry_ptr pGeo;
    sple_linestring_ptr line = NULL;
    sple_point_ptr point;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    *size = 0;
    *result = NULL;
    if (n_points != 0 || n_polygons != 0)
	return;
    if (n_linestrings != 1)
	return;
    if (vertex < 0)
	index = line->points;	/* resolvind negative point index for EndPoint() request */
    else
	index = vertex;
    if (index > 0 && index <= line->points)
	;
    else
	return;
/* builds a new POINT GEOMETRY */
    point =
	sple_alloc_point ((line->geometry + (index - 1))->x,
			  (line->geometry + (index - 1))->y);
    new_geom = sple_alloc_geometry ();
    new_geom->entities = 1;
    new_geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(new_geom->geometries) =
	sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
    sple_blob_encode (new_geom, result, size);	/* generating BLOB internal representation */
    sple_free_geometry (new_geom);
}

SPATIALITE_DECLARE int
sple_num_interior_rings (sple_geometry_ptr geom, int *borders)
{
/* returns the number of interior rings from a POLYGON geometry */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_polygon_ptr polyg = NULL;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		n_polygons++;
		break;
	    };
      }
    if (n_points != 0 || n_linestrings != 0)
	return 0;
    if (n_polygons != 1)
	return 0;
    *borders = polyg->borders;
    return 1;
}

SPATIALITE_DECLARE void
sple_ring_n (sple_geometry_ptr geom, int border,
	     unsigned char **result, int *size)
{
/* extracts the Nth internal border from a POLYGON geometry */
    int ie;
    int iv;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_polygon_ptr polyg = NULL;
    sple_linestring_ptr line;
    sple_linestring_ptr new_line;
    sple_geometry_ptr new_geom;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		n_polygons++;
		break;
	    };
      }
    *size = 0;
    *result = NULL;
    if (n_points != 0 || n_linestrings != 0)
	return;
    if (n_polygons != 1)
	return;
    if (border < 0)
	;			/* ExteriorRing() SQL function defines a negative ring index */
    else if (border > 0 && border <= polyg->borders)
	;
    else
	return;
/* builds a new LINESTRING GEOMETRY */
    if (border < 0)
	line = polyg->external;	/* ExteriorRing() SQL function defines a negative ring index */
    else
	line = *(polyg->internals + (border - 1));
    new_line = sple_alloc_linestring (line->points);
    for (iv = 0; iv < line->points; iv++)
	sple_set_point (new_line, iv, (line->geometry + iv)->x,
			(line->geometry + iv)->y);
    new_geom = sple_alloc_geometry ();
    new_geom->entities = 1;
    new_geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(new_geom->geometries) =
	sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) new_line);
    sple_blob_encode (new_geom, result, size);	/* generating BLOB internal representation */
    sple_free_geometry (new_geom);
}

SPATIALITE_DECLARE int
sple_num_geometries (sple_geometry_ptr geom, int *geoms)
{
/* returns the number of geometries from a GEOMETRYCOLLECTION or MULTIxxxx geometry */
    if (geom->entities >= 1)
      {
	  *geoms = geom->entities;
	  return 1;
      }
    else
      {
	  *geoms = 0;
	  return 0;
      }
}

SPATIALITE_DECLARE void
sple_geometry_n (sple_geometry_ptr geom, int entity, unsigned char **result,
		 int *size)
{
/* extracts the Nth geometry from a GEOMETRYCOLLECTION or MULTIxxxx geometry */
    int ib;
    int iv;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point = NULL;
    sple_polygon_ptr polyg = NULL;
    sple_linestring_ptr line = NULL;
    sple_geometry_ptr new_geom;
    sple_point_ptr new_point;
    sple_polygon_ptr new_polyg;
    sple_linestring_ptr new_line;
    *size = 0;
    *result = NULL;
    if (geom->entities < 1)
	return;
    if (entity >= 1 && entity <= geom->entities)
	;
    else
	return;
    pGeo = *(geom->geometries + (entity - 1));
    switch (pGeo->type)
      {
      case SPLE_POINT:
	  point = (sple_point_ptr) (pGeo->geometry);
	  break;
      case SPLE_LINESTRING:
	  line = (sple_linestring_ptr) (pGeo->geometry);
	  break;
      case SPLE_POLYGON:
	  polyg = (sple_polygon_ptr) (pGeo->geometry);
	  break;
      };
    if (point)
      {
	  /* builds a new POINT GEOMETRY */
	  new_point = sple_alloc_point (point->x, point->y);
	  new_geom = sple_alloc_geometry ();
	  new_geom->entities = 1;
	  new_geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
	  *(new_geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
	  sple_blob_encode (new_geom, result, size);	/* generating BLOB internal representation */
	  sple_free_geometry (new_geom);
      }
    if (line)
      {
	  /* builds a new LINESTRING GEOMETRY */
	  new_line = sple_alloc_linestring (line->points);
	  for (iv = 0; iv < line->points; iv++)
	      sple_set_point (new_line, iv, (line->geometry + iv)->x,
			      (line->geometry + iv)->y);
	  new_geom = sple_alloc_geometry ();
	  new_geom->entities = 1;
	  new_geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
	  *(new_geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) new_line);
	  sple_blob_encode (new_geom, result, size);	/* generating BLOB internal representation */
	  sple_free_geometry (new_geom);
      }
    if (polyg)
      {
	  /* builds a new POLYGON GEOMETRY */
	  line = polyg->external;
	  new_polyg = sple_alloc_polygon (line->points, polyg->borders);
	  new_line = new_polyg->external;
	  for (iv = 0; iv < line->points; iv++)
	      sple_set_point (new_line, iv, (line->geometry + iv)->x,
			      (line->geometry + iv)->y);
	  for (ib = 0; ib < polyg->borders; ib++)
	    {
		line = *(polyg->internals + ib);
		new_line =
		    sple_alloc_internal_border (new_polyg, ib, line->points);
		for (iv = 0; iv < line->points; iv++)
		    sple_set_point (new_line, iv, (line->geometry + iv)->x,
				    (line->geometry + iv)->y);
	    }
	  new_geom = sple_alloc_geometry ();
	  new_geom->entities = 1;
	  new_geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
	  *(new_geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_POLYGON, (void *) new_polyg);
	  sple_blob_encode (new_geom, result, size);	/* generating BLOB internal representation */
	  sple_free_geometry (new_geom);
      }
}

SPATIALITE_DECLARE int
sple_count_geometries (sple_geometry_ptr geom)
{
/* returns the number of geometries from a  geometry */
    return geom->entities;
}

SPATIALITE_DECLARE int
sple_length (sple_geometry_ptr geom, double *xlength)
{
/* 
/ returns the length for a LINESTRING geometry,
/ or the total length for a MULTILINESTRING geometry
 */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_linestring_ptr line = NULL;
    double tot_length = 0.0;
    double length;
    double xx1 = 0.0;
    double xx2;
    double yy1 = 0.0;
    double yy2;
    double x;
    double y;
    double dist;
    int iv;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		length = 0.0;
		for (iv = 0; iv < line->points; iv++)
		  {
		      if (iv == 0)
			{
			    xx1 = (line->geometry + iv)->x;
			    yy1 = (line->geometry + iv)->y;
			}
		      else
			{
			    xx2 = (line->geometry + iv)->x;
			    yy2 = (line->geometry + iv)->y;
			    x = xx1 - xx2;
			    y = yy1 - yy2;
			    dist = sqrt ((x * x) + (y * y));
			    length += dist;
			    xx1 = xx2;
			    yy1 = yy2;
			}
		  }
		tot_length += length;
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    if (n_points != 0 || n_polygons != 0)
	return 0;
    if (n_linestrings == 0)
	return 0;
    *xlength = tot_length;
    return 1;
}

SPATIALITE_DECLARE int
sple_area (sple_geometry_ptr geom, double *xarea)
{
/* 
/ returns the area for a POLYGON geometry,
/ or the total area for a MULTIPOLYGON geometry
 */
    int ie;
    int ib;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_polygon_ptr polyg = NULL;
    sple_linestring_ptr line;
    double tot_area = 0.0;
    double area;
    double xx;
    double yy;
    double x;
    double y;
    int iv;
    int ix;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		line = polyg->external;
		area = 0.0;
		for (iv = 0; iv < line->points; iv++)
		  {
		      /* computes the area for exterior ring [POSITIVE] */
		      xx = (line->geometry + iv)->x;
		      yy = (line->geometry + iv)->y;
		      ix = (iv + 1) % line->points;
		      x = (line->geometry + ix)->x;
		      y = (line->geometry + ix)->y;
		      area += ((x * yy) - (xx * y));
		  }
		area /= 2.0;
		tot_area += fabs (area);
		for (ib = 0; ib < polyg->borders; ib++)
		  {
		      line = *(polyg->internals + ib);
		      area = 0.0;
		      for (iv = 0; iv < line->points; iv++)
			{
			    /* computes the area for interior ring [NEGATIVE] */
			    xx = (line->geometry + iv)->x;
			    yy = (line->geometry + iv)->y;
			    ix = (iv + 1) % line->points;
			    x = (line->geometry + ix)->x;
			    y = (line->geometry + ix)->y;
			    area += ((x * yy) - (xx * y));
			}
		      area /= 2.0;
		      tot_area -= fabs (area);
		  }
		n_polygons++;
		break;
	    };
      }
    if (n_points != 0 || n_linestrings != 0)
	return 0;
    if (n_polygons == 0)
	return 0;
    *xarea = tot_area;
    return 1;
}

SPATIALITE_DECLARE void
sple_centroid (sple_geometry_ptr geom, unsigned char **result, int *size)
{
/* 
/ returns the centroid for a POLYGON geometry,
/ or the centroid of the max-area polygon inside a MULTIPOLYGON geometry
 */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    sple_polygon_ptr sv_polyg = NULL;
    sple_polygon_ptr polyg = NULL;
    sple_linestring_ptr line;
    sple_geometry_ptr new_geom;
    sple_point_ptr point;
    double max_area = 0.0;
    double area;
    double xx;
    double yy;
    double x;
    double y;
    int iv;
    int ix;
    double coeff;
    double cx = 0.0;
    double cy = 0.0;
    double rx;
    double ry;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  /* pass I - we try to find the maximun area POLYGON inside this geometry */
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		line = polyg->external;
		area = 0.0;
		for (iv = 0; iv < line->points; iv++)
		  {
		      xx = (line->geometry + iv)->x;
		      yy = (line->geometry + iv)->y;
		      ix = (iv + 1) % line->points;
		      x = (line->geometry + ix)->x;
		      y = (line->geometry + ix)->y;
		      area += ((x * yy) - (xx * y));
		  }
		area /= 2.0;
		area = fabs (area);
		if (area > max_area)
		  {
		      max_area = area;
		      sv_polyg = polyg;
		  }
		n_polygons++;
		break;
	    };
      }
    *size = 0;
    *result = NULL;
    if (n_points != 0 || n_linestrings != 0)
	return;
    if (n_polygons == 0)
	return;
/*
/ now we surely have:
/ or the unique POLYGON within this geometry
/ either the maximun area POLYGON if that one is a MULTIPOLYGON
*/
    coeff = 1.0 / (max_area * 6.0);
    line = sv_polyg->external;
    for (iv = 0; iv < line->points; iv++)
      {
	  xx = (line->geometry + iv)->x;
	  yy = (line->geometry + iv)->y;
	  ix = (iv + 1) % line->points;
	  x = (line->geometry + ix)->x;
	  y = (line->geometry + ix)->y;
	  cx += (xx + x) * ((x * yy) - (xx * y));
	  cy += (yy + y) * ((x * yy) - (xx * y));
      }
    rx = fabs (cx * coeff);
    ry = fabs (cy * coeff);
/* builds a new POINT GEOMETRY */
    point = sple_alloc_point (rx, ry);
    new_geom = sple_alloc_geometry ();
    new_geom->entities = 1;
    new_geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(new_geom->geometries) =
	sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
    sple_blob_encode (new_geom, result, size);	/* generating BLOB internal representation */
    sple_free_geometry (new_geom);
}

SPATIALITE_DECLARE void
sple_mbr_geometry (sple_geometry_ptr geom)
{
/* computes the MBR for this geometry */
    int ie;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point = NULL;
    sple_linestring_ptr line = NULL;
    sple_polygon_ptr polyg = NULL;
    geom->minx = DBL_MAX;
    geom->miny = DBL_MAX;
    geom->maxx = DBL_MIN;
    geom->maxy = DBL_MIN;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		if (point->x < geom->minx)
		    geom->minx = point->x;
		if (point->y < geom->miny)
		    geom->miny = point->y;
		if (point->x > geom->maxx)
		    geom->maxx = point->x;
		if (point->y > geom->maxy)
		    geom->maxy = point->y;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		sple_mbr_linestring (line);
		if (line->minx < geom->minx)
		    geom->minx = line->minx;
		if (line->miny < geom->miny)
		    geom->miny = line->miny;
		if (line->minx > geom->maxx)
		    geom->maxx = line->maxx;
		if (line->miny > geom->maxy)
		    geom->maxy = line->maxy;
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		sple_mbr_polygon (polyg);
		if (polyg->minx < geom->minx)
		    geom->minx = polyg->minx;
		if (polyg->miny < geom->miny)
		    geom->miny = polyg->miny;
		if (polyg->minx > geom->maxx)
		    geom->maxx = polyg->maxx;
		if (polyg->miny > geom->maxy)
		    geom->maxy = polyg->maxy;
		break;
	    };
      }
}

SPATIALITE_DECLARE void
sple_mbr_linestring (sple_linestring_ptr line)
{
/* computes the MBR for this linestring */
    int iv;
    double x;
    double y;
    line->minx = DBL_MAX;
    line->miny = DBL_MAX;
    line->maxx = DBL_MIN;
    line->maxy = DBL_MIN;
    for (iv = 0; iv < line->points; iv++)
      {
	  x = (line->geometry + iv)->x;
	  y = (line->geometry + iv)->y;
	  if (x < line->minx)
	      line->minx = x;
	  if (y < line->miny)
	      line->miny = y;
	  if (x > line->maxx)
	      line->maxx = x;
	  if (y > line->maxy)
	      line->maxy = y;
      }
}

SPATIALITE_DECLARE void
sple_mbr_polygon (sple_polygon_ptr polyg)
{
/* computes the MBR for this polygon */
    sple_linestring_ptr line;
    polyg->minx = DBL_MAX;
    polyg->miny = DBL_MAX;
    polyg->maxx = DBL_MIN;
    polyg->maxy = DBL_MIN;
    line = polyg->external;
    sple_mbr_linestring (line);
    if (line->minx < polyg->minx)
	polyg->minx = line->minx;
    if (line->miny < polyg->miny)
	polyg->miny = line->miny;
    if (line->minx > polyg->maxx)
	polyg->maxx = line->maxx;
    if (line->miny > polyg->maxy)
	polyg->maxy = line->maxy;
}

SPATIALITE_DECLARE void
sple_blob_to_wkb (sple_geometry_ptr geom, unsigned char **result, int *size)
{
/* extracts the WKB representation from an internal GEOMETRY */
    int ie;
    int ib;
    int iv;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    int type;
    unsigned char *ptr;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point;
    sple_polygon_ptr polyg;
    sple_linestring_ptr line;
    *size = 0;
    *result = NULL;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  /* checks the entities */
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    if (n_points == 0 && n_polygons == 0 && n_linestrings == 0)
	return;
    *size = 5;
    type = SPLE_GEOMETRYCOLLECTION;
    if (n_points == 1 && n_polygons == 0 && n_linestrings == 0)
	type = SPLE_POINT;
    if (n_points > 1 && n_polygons == 0 && n_linestrings == 0)
	type = SPLE_MULTIPOINT;
    if (n_points == 0 && n_polygons == 0 && n_linestrings == 1)
	type = SPLE_LINESTRING;
    if (n_points == 0 && n_polygons == 0 && n_linestrings > 1)
	type = SPLE_MULTILINESTRING;
    if (n_points == 0 && n_polygons == 1 && n_linestrings == 0)
	type = SPLE_POLYGON;
    if (n_points == 0 && n_polygons > 1 && n_linestrings == 0)
	type = SPLE_MULTIPOLYGON;
    if (type == SPLE_MULTIPOINT || type == SPLE_MULTILINESTRING ||
	type == SPLE_MULTIPOLYGON || type == SPLE_GEOMETRYCOLLECTION)
	*size += 4;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  /* computing the WKB size */
	  pGeo = *(geom->geometries + ie);
	  if (type == SPLE_MULTIPOINT || type == SPLE_MULTILINESTRING ||
	      type == SPLE_MULTIPOLYGON || type == SPLE_GEOMETRYCOLLECTION)
	      *size += 5;
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		*size += sizeof (double) * 2;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		*size += 4 + ((sizeof (double) * 2) * line->points);
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		line = polyg->external;
		*size += 8 + ((sizeof (double) * 2) * line->points);
		for (ib = 0; ib < polyg->borders; ib++)
		  {
		      line = *(polyg->internals + ib);
		      *size += 4 + ((sizeof (double) * 2) * line->points);
		  }
		break;
	    };
      }
    *result = malloc (*size);
    ptr = *result;
    *ptr = 0x01;		/* little endian byte order */
    sple_export_32 (ptr + 1, type, 1);	/* the main CLASS TYPE */
    ptr += 5;
    if (type == SPLE_MULTIPOINT || type == SPLE_MULTILINESTRING ||
	type == SPLE_MULTIPOLYGON || type == SPLE_GEOMETRYCOLLECTION)
      {
	  sple_export_32 (ptr, geom->entities, 1);	/* it's a collection; # entities */
	  ptr += 4;
      }
    for (ie = 0; ie < geom->entities; ie++)
      {
	  /* builds the WKB */
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		if (type == SPLE_MULTIPOINT)
		  {
		      *ptr = 0x01;
		      sple_export_32 (ptr + 1, SPLE_POINT, 1);	/* it's a collection: the CLASS TYPE for this element */
		      ptr += 5;
		  }
		sple_export_64 (ptr, point->x, 1);	/* X */
		ptr += 8;
		sple_export_64 (ptr, point->y, 1);	/* Y */
		ptr += 8;
		break;
	    case SPLE_LINESTRING:
		if (type == SPLE_MULTILINESTRING)
		  {
		      *ptr = 0x01;
		      sple_export_32 (ptr + 1, SPLE_LINESTRING, 1);	/* it's a collection: the CLASS TYPE for this element */
		      ptr += 5;
		  }
		line = (sple_linestring_ptr) (pGeo->geometry);
		sple_export_32 (ptr, line->points, 1);	/* # points */
		ptr += 4;
		for (iv = 0; iv < line->points; iv++)
		  {
		      sple_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X */
		      ptr += 8;
		      sple_export_64 (ptr, (line->geometry + iv)->y, 1);	/* Y */
		      ptr += 8;
		  }
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		if (type == SPLE_MULTIPOLYGON)
		  {
		      *ptr = 0x01;
		      sple_export_32 (ptr + 1, SPLE_POLYGON, 1);	/* it's a collection: the CLASS TYPE for this element */
		      ptr += 5;
		  }
		sple_export_32 (ptr, polyg->borders + 1, 1);	/* # rings */
		ptr += 4;
		line = polyg->external;
		sple_export_32 (ptr, line->points, 1);	/* # points - exterior ring */
		ptr += 4;
		for (iv = 0; iv < line->points; iv++)
		  {
		      sple_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - exterior ring */
		      ptr += 8;
		      sple_export_64 (ptr, (line->geometry + iv)->y, 1);	/* Y - exterior ring */
		      ptr += 8;
		  }
		for (ib = 0; ib < polyg->borders; ib++)
		  {
		      line = *(polyg->internals + ib);
		      sple_export_32 (ptr, line->points, 1);	/* points - interior ring */
		      ptr += 4;
		      for (iv = 0; iv < line->points; iv++)
			{
			    sple_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - interior ring */
			    ptr += 8;
			    sple_export_64 (ptr, (line->geometry + iv)->y, 1);	/* Y - interior ring */
			    ptr += 8;
			}
		  }
		break;
	    };
      }
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_parse_wkb (const unsigned char *wkb, int size, short type)
{
/* tryes to build a GEOMETRY by parsing a WKB encoded BLOB */
    int little_endian;
    int wkb_type;
    int sub_type;
    int entities;
    int borders;
    int points;
    int x_size;
    int ie;
    int ib;
    int iv;
    double x;
    double y;
    sple_point_ptr point;
    sple_linestring_ptr line;
    sple_polygon_ptr polyg;
    const unsigned char *ptr;
    sple_geometry_ptr geom = NULL;
    if (size < 5)
	return NULL;		/* too short to be a WKB */
    if (*wkb == SPLE_LITTLE_ENDIAN)
	little_endian = 1;
    else if (*wkb == SPLE_BIG_ENDIAN)
	little_endian = 0;
    else
	return NULL;		/* illegal byte ordering; neither BIG-ENDIAN nor LITTLE-ENDIAN */
    wkb_type = sple_import_32 (wkb + 1, little_endian);
    if (wkb_type == SPLE_POINT || wkb_type == SPLE_LINESTRING
	|| wkb_type == SPLE_POLYGON || wkb_type == SPLE_MULTIPOINT
	|| wkb_type == SPLE_MULTILINESTRING || wkb_type == SPLE_MULTIPOLYGON
	|| wkb_type == SPLE_GEOMETRYCOLLECTION)
	;
    else
	return NULL;		/* illegal GEOMETRY CLASS */
    if (type < 0)
	;			/* no restrinction about GEOMETRY CLASS TYPE */
    else
      {
	  if (wkb_type != type)
	      return NULL;	/* invalid CLASS TYPE for request */
      }
    x_size = 5;			/* standard WKB header */
    switch (wkb_type)
      {
	  /* computes # entities and expected WKB size */
      case SPLE_POINT:
	  entities = 1;
	  x_size += (sizeof (double) * 2);	/* WKB POINT size */
	  break;
      case SPLE_LINESTRING:
	  entities = 1;
	  if (size < 9)
	      return NULL;	/* WKB too short */
	  points = sple_import_32 (wkb + 5, little_endian);	/* # points */
	  x_size += 4 + ((sizeof (double) * 2) * points);	/* size for array of # points POINTS */
	  break;
      case SPLE_POLYGON:
	  entities = 1;
	  if (size < 9)
	      return NULL;	/* WKB too short */
	  borders = sple_import_32 (wkb + 5, little_endian);	/* # borders */
	  x_size = 9;
	  ptr = wkb + 9;
	  for (ib = 0; ib < borders; ib++)
	    {
		if (size < ((ptr - wkb) + 4))
		    return NULL;	/* WKB too short */
		points = sple_import_32 (ptr, little_endian);	/* # points in this ring */
		x_size += 4 + ((sizeof (double) * 2) * points);	/* size for array of # points POINTS */
		ptr += 4 + ((sizeof (double) * 2) * points);
	    }
	  break;
      default:
	  if (size < 9)
	      return NULL;	/* WKB too short */
	  entities = sple_import_32 (wkb + 5, little_endian);	/* # entities inside COLLECTION */
	  ptr = wkb + 9;
	  x_size = 9;
	  for (ie = 0; ie < entities; ie++)
	    {
		if (size < ((ptr - wkb) + 5))
		    return NULL;	/* WKB too short */
		sub_type = sple_import_32 (ptr + 1, little_endian);	/* GEOMETRY CLASS for this entity */
		ptr += 5;
		x_size += 5;	/* geometry class for entity */
		switch (sub_type)
		  {
		  case SPLE_POINT:
		      x_size += (sizeof (double) * 2);	/* WKB POINT size */
		      ptr += (sizeof (double) * 2);
		      break;
		  case SPLE_LINESTRING:
		      if (size < ((ptr - wkb) + 4))
			  return NULL;	/* WKB too short */
		      points = sple_import_32 (ptr, little_endian);	/* # points */
		      x_size += 4 + ((sizeof (double) * 2) * points);	/* size for array of # points POINTS */
		      ptr += 4 + ((sizeof (double) * 2) * points);
		      break;
		  case SPLE_POLYGON:
		      if (size < ((ptr - wkb) + 4))
			  return NULL;	/* WKB too short */
		      borders = sple_import_32 (ptr, little_endian);	/* # borders */
		      x_size += 4;
		      ptr += 4;
		      for (ib = 0; ib < borders; ib++)
			{
			    if (size < ((ptr - wkb) + 4))
				return NULL;	/* WKB too short */
			    points = sple_import_32 (ptr, little_endian);	/* # points in this ring */
			    x_size += 4 + ((sizeof (double) * 2) * points);	/* size for array of # points POINTS */
			    ptr += 4 + ((sizeof (double) * 2) * points);
			}
		      break;
		  };
	    }
	  break;
      };
    if (x_size != size)
	return NULL;
/* allocates the GEOMETRY to be returned */
    ptr = wkb + 5;
    geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = entities;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr) * entities);
    switch (wkb_type)
      {
	  /* decodes entities from WKB and inserts them into GEOMETRY to be returned  */
      case SPLE_POINT:
	  x = sple_import_64 (ptr, little_endian);	/* X */
	  y = sple_import_64 (ptr + 8, little_endian);	/* Y */
	  point = sple_alloc_point (x, y);
	  *(geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
	  break;
      case SPLE_LINESTRING:
	  points = sple_import_32 (ptr, little_endian);	/* # points */
	  ptr += 4;
	  line = sple_alloc_linestring (points);
	  for (iv = 0; iv < points; iv++)
	    {
		x = sple_import_64 (ptr, little_endian);	/* X */
		y = sple_import_64 (ptr + 8, little_endian);	/* Y */
		ptr += 16;
		sple_set_point (line, iv, x, y);
	    }
	  *(geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) line);
	  break;
      case SPLE_POLYGON:
	  entities = 1;
	  borders = sple_import_32 (ptr, little_endian);	/* # borders */
	  ptr += 4;
	  points = sple_import_32 (ptr, little_endian);	/* # points - exterior ring */
	  ptr += 4;
	  polyg = sple_alloc_polygon (points, borders - 1);
	  line = polyg->external;
	  for (iv = 0; iv < points; iv++)
	    {
		x = sple_import_64 (ptr, little_endian);	/* X - exterior ring */
		y = sple_import_64 (ptr + 8, little_endian);	/* Y - exterior ring */
		ptr += 16;
		sple_set_point (line, iv, x, y);
	    }
	  for (ib = 0; ib < (borders - 1); ib++)
	    {
		points = sple_import_32 (ptr, little_endian);	/* # points - interior ring */
		ptr += 4;
		line = sple_alloc_internal_border (polyg, ib, points);
		for (iv = 0; iv < points; iv++)
		  {
		      x = sple_import_64 (ptr, little_endian);	/* X - interior ring */
		      y = sple_import_64 (ptr + 8, little_endian);	/* Y - interior ring */
		      ptr += 16;
		      sple_set_point (line, iv, x, y);
		  }
	    }
	  *(geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_POLYGON, (void *) polyg);
	  break;
      default:
	  ptr = wkb + 9;
	  for (ie = 0; ie < entities; ie++)
	    {
		/* loops on each entity in the COLLECTION */
		sub_type = sple_import_32 (ptr + 1, little_endian);	/* GEOMETRY CLASS for this entity */
		ptr += 5;
		switch (sub_type)
		  {
		  case SPLE_POINT:
		      x = sple_import_64 (ptr, little_endian);	/* X */
		      y = sple_import_64 (ptr + 8, little_endian);	/* Y */
		      ptr += 16;
		      point = sple_alloc_point (x, y);
		      *(geom->geometries + ie) =
			  sple_alloc_generic_geometry (SPLE_POINT,
						       (void *) point);
		      break;
		  case SPLE_LINESTRING:
		      points = sple_import_32 (ptr, little_endian);	/* # points */
		      ptr += 4;
		      line = sple_alloc_linestring (points);
		      for (iv = 0; iv < points; iv++)
			{
			    x = sple_import_64 (ptr, little_endian);	/* X */
			    y = sple_import_64 (ptr + 8, little_endian);	/* Y */
			    ptr += 16;
			    sple_set_point (line, iv, x, y);
			}
		      *(geom->geometries + ie) =
			  sple_alloc_generic_geometry (SPLE_LINESTRING,
						       (void *) line);
		      break;
		  case SPLE_POLYGON:
		      borders = sple_import_32 (ptr, little_endian);	/* # borders */
		      ptr += 4;
		      points = sple_import_32 (ptr, little_endian);	/* # points - exterior ring */
		      ptr += 4;
		      polyg = sple_alloc_polygon (points, borders - 1);
		      line = polyg->external;
		      for (iv = 0; iv < points; iv++)
			{
			    x = sple_import_64 (ptr, little_endian);	/* X - exterior ring */
			    y = sple_import_64 (ptr + 8, little_endian);	/* Y - exterior ring */
			    ptr += 16;
			    sple_set_point (line, iv, x, y);
			}
		      for (ib = 0; ib < (borders - 1); ib++)
			{
			    points = sple_import_32 (ptr, little_endian);	/* # points - interior ring */
			    ptr += 4;
			    line =
				sple_alloc_internal_border (polyg, ib, points);
			    for (iv = 0; iv < points; iv++)
			      {
				  x = sple_import_64 (ptr, little_endian);	/* X - interior ring */
				  y = sple_import_64 (ptr + 8, little_endian);	/* Y - interior ring */
				  ptr += 16;
				  sple_set_point (line, iv, x, y);
			      }
			}
		      *(geom->geometries + ie) =
			  sple_alloc_generic_geometry (SPLE_POLYGON,
						       (void *) polyg);
		      break;
		  };
	    }
	  break;
      };
    return geom;
}
