/*
/
/ sple_shapefile.c - SpatiaLite - spatial extension for SQLite - shapefile support
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2007  Alessandro Furieri
/
/ This library is free software; you can redistribute it and/or
/ modify it under the terms of the GNU Lesser General Public
/ License as published by the Free Software Foundation; either
/ version 2.1 of the License, or (at your option) any later version.
/
/ This library is distributed in the hope that it will be useful,
/ but WITHOUT ANY WARRANTY; without even the implied warranty of
/ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/ Lesser General Public License for more details.
/
/ You should have received a copy of the GNU Lesser General Public
/ License along with this library; if not, write to the Free Software
/ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA/
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include "sqlite3.h"
#include "spatialite.h"

static int create_table (const unsigned char *table,
			 const unsigned char *column, sple_dbf_list_ptr list,
			 char **msg);
static int is_duplicate_column (sple_dbf_list_ptr list, char *name);
static int is_reserved_sql_name (char *name);
static int is_reserved_sqlite_name (char *name);
static sple_geometry_ptr new_geom_point (sple_point_ptr point);
static sple_geometry_ptr new_geom_line (sple_linestring_ptr line);
static void add_line_to_geom (sple_geometry_ptr geom, sple_linestring_ptr line);
static sple_geometry_ptr new_geom_polyg (sple_linestring_ptr ring,
					 sple_polygon_ptr * pg);
static void add_polyg_to_geom (sple_geometry_ptr geom, sple_linestring_ptr ring,
			       sple_polygon_ptr * pg);
static void add_ring_to_polyg (sple_polygon_ptr polyg,
			       sple_linestring_ptr ring);
static int is_clockwise_ring (sple_linestring_ptr ring);
static void sane_clockwise (sple_polygon_ptr polyg);
static void clean_sql_string (char *string);
static int is_date (const unsigned char *value, int len);

SPATIALITE_DECLARE int
sple_load_shapefile (const unsigned char *path, const unsigned char *table,
		     const unsigned char *column, int srid, int *entities,
		     int *xshape, char **error)
{
/* trying to load an external shapefile into an SQLite table */
    extern sqlite3 *this_db;
    FILE *fl_shx = NULL;
    FILE *fl_shp = NULL;
    FILE *fl_dbf = NULL;
    char sql[8192];
    char xpath[1024];
    char value[1024];
    int ret;
    char **result;
    int nrow;
    int ncolumn;
    char *errMsg = NULL;
    int len;
    int i;
    int table_already_exists = 0;
    unsigned char *buf_shp = NULL;
    int buf_size = 1024;
    int rd;
    long shape;
    long xxshape;
    unsigned char buf_shx[256];
    long size_shp;
    long size_shx;
    int off_dbf;
    long cur_dbf = 0;
    unsigned char bf[1024];
    unsigned char *buf_dbf = NULL;
    int dbf_size;
    int dbf_reclen = 0;
    int dbf_recno;
    int ind;
    char field_name[16];
    sple_dbf_list_ptr dbf_list = NULL;
    double x;
    double y;
    long n;
    long n1;
    int base;
    int start;
    int end;
    int points;
    long skpos;
    long sz;
    long off_shp;
    long sz_shp;
    long cur_size;
    int iv;
    int rows = 0;
    sple_geometry_ptr geom;
    sple_point_ptr point = NULL;
    sple_linestring_ptr line = NULL;
    sple_polygon_ptr polyg = NULL;
    char *err_sql = NULL;
    sple_dbf_field_ptr dbf_field;
    char date[16];
    char fmt[16];
    char dummy[64];
    unsigned char *wkb_buf;
    int wkb_size;
    unsigned char *real_sql;
    int sql_sz;
    int i_sql;
    char *sys_err;
    *entities = 0;
    *xshape = 0;
/* pass I - checks if SQLite table already exists */
    strcpy (sql, "SELECT name FROM sqlite_master WHERE type = 'table'");
    ret = sqlite3_get_table (this_db, sql, &result, &nrow, &ncolumn, &errMsg);
    if (ret != SQLITE_OK)
      {
	  len = strlen (errMsg);
	  *error = malloc (len + 1);
	  strcpy (*error, errMsg);
	  sqlite3_free (errMsg);
	  return -1;
      }
    for (i = 0; i < nrow; i++)
      {
	  if (strcasecmp (result[i], (char *) table) == 0)
	    {
		table_already_exists = 1;
		break;
	    }
      }
    sqlite3_free_table (result);
    if (table_already_exists)
      {
	  sprintf (sql, "SQLite table '%s' already exists", table);
	  len = strlen (sql);
	  *error = malloc (len + 1);
	  strcpy (*error, sql);
	  return -1;
      }
/* pass II - checks if shapefile exists and is accessible */
    sprintf (xpath, "%s.shx", path);
    fl_shx = fopen (xpath, "rb");
    if (!fl_shx)
	goto no_file;
    sprintf (xpath, "%s.shp", path);
    fl_shp = fopen (xpath, "rb");
    if (!fl_shp)
	goto no_file;
    sprintf (xpath, "%s.dbf", path);
    fl_dbf = fopen (xpath, "rb");
    if (!fl_dbf)
	goto no_file;
/* reading SHX file header */
    rd = fread (buf_shx, sizeof (unsigned char), 100, fl_shx);
    if (rd != 100)
	goto error;
    if (sple_import_32 (buf_shx + 0, SPLE_BIG_ENDIAN) != 9994)	/* checks the SHX magic number */
	goto error;
    size_shx = sple_import_32 (buf_shx + 24, SPLE_BIG_ENDIAN);
/* reading SHP file header */
    buf_shp = malloc (sizeof (unsigned char) * buf_size);
    rd = fread (buf_shp, sizeof (unsigned char), 100, fl_shp);
    if (rd != 100)
	goto error;
    if (sple_import_32 (buf_shp + 0, SPLE_BIG_ENDIAN) != 9994)	/* checks the SHP magic number */
	goto error;
    size_shp = sple_import_32 (buf_shp + 24, SPLE_BIG_ENDIAN);
    xxshape = sple_import_32 (buf_shp + 32, SPLE_LITTLE_ENDIAN);
    *xshape = xxshape;
    if (xxshape == 1 || xxshape == 11 || xxshape == 21 ||
	xxshape == 3 || xxshape == 13 || xxshape == 23 ||
	xxshape == 5 || xxshape == 15 || xxshape == 25)
	;
    else
	goto unsupported;
/* reading DBF file header */
    rd = fread (bf, sizeof (unsigned char), 32, fl_dbf);
    if (rd != 32)
	goto error;
    if (*bf != 0x03)		/* checks the DBF magic number */
	goto error;
    dbf_recno = sple_import_32 (bf + 4, SPLE_LITTLE_ENDIAN);
    dbf_size = sple_import_16 (bf + 8, SPLE_LITTLE_ENDIAN);
    dbf_reclen = sple_import_16 (bf + 10, SPLE_LITTLE_ENDIAN);
    dbf_size--;
    off_dbf = 0;
    for (ind = 32; ind < dbf_size; ind += 32)
      {
	  /* fetches DBF fields definitions */
	  rd = fread (bf, sizeof (unsigned char), 32, fl_dbf);
	  if (rd != 32)
	      goto error;
	  memcpy (field_name, bf, 11);
	  field_name[11] = '\0';
	  if (!dbf_list)
	      dbf_list = sple_alloc_dbf_list ("PK_UID", (char *) column);
	  sple_add_dbf_field (dbf_list, field_name, *(bf + 11), off_dbf,
			      *(bf + 16), *(bf + 17));
	  off_dbf += *(bf + 16);
      }
/* creating the table */
    if (!create_table (table, column, dbf_list, &err_sql))
	goto no_table;
/* allocating DBF buffer */
    buf_dbf = malloc (sizeof (unsigned char) * dbf_reclen);
    cur_dbf = dbf_size + 1;
    cur_size = 50;
    while (1)
      {
/* reading SHX entities - index */
	  geom = NULL;
	  if (cur_size >= size_shx)
	      break;
	  rd = fread (buf_shx, sizeof (unsigned char), 8, fl_shx);
	  if (rd != 8)
	      goto error;
	  cur_size += 4;
	  (*entities)++;
	  off_shp = sple_import_32 (buf_shx, SPLE_BIG_ENDIAN);
	  sz_shp = sple_import_32 (buf_shx + 4, SPLE_BIG_ENDIAN);
/* reading corresponding SHP entity - geometry */
	  skpos = fseek (fl_shp, off_shp * 2, SEEK_SET);
	  if (skpos != 0)
	      goto error;
	  rd = fread (buf_shp, sizeof (unsigned char), 12, fl_shp);
	  if (rd != 12)
	      goto error;
	  sz = sple_import_32 (buf_shp + 4, SPLE_BIG_ENDIAN);
	  shape = sple_import_32 (buf_shp + 8, SPLE_LITTLE_ENDIAN);
	  if (shape != xxshape)
	      goto error;
	  if ((sz * 2) > buf_size)
	    {
		/* current buffer is too small; we need to allocate a bigger buffer */
		free (buf_shp);
		buf_size = sz * 2;
		buf_shp = malloc (sizeof (unsigned char) * buf_size);
	    }
	  if (shape == 1 || shape == 11 || shape == 21)
	    {
		/* shape point */
		rd = fread (buf_shp, sizeof (unsigned char), 16, fl_shp);
		if (rd != 16)
		    goto error;
		x = sple_import_64 (buf_shp, SPLE_LITTLE_ENDIAN);
		y = sple_import_64 (buf_shp + 8, SPLE_LITTLE_ENDIAN);
		point = sple_alloc_point (x, y);
		geom = new_geom_point (point);
	    }
	  if (shape == 3 || shape == 13 || shape == 23)
	    {
		/* shape polyline */
		rd = fread (buf_shp, sizeof (unsigned char), 32, fl_shp);
		if (rd != 32)
		    goto error;
		rd = fread (buf_shp, sizeof (unsigned char),
			    (sz * 2) - 36, fl_shp);
		if (rd != (sz * 2) - 36)
		    goto error;
		n = sple_import_32 (buf_shp, SPLE_LITTLE_ENDIAN);
		n1 = sple_import_32 (buf_shp + 4, SPLE_LITTLE_ENDIAN);
		base = 8 + (n * 4);
		start = 0;
		for (ind = 0; ind < n; ind++)
		  {
		      if (ind < (n - 1))
			  end =
			      sple_import_32 (buf_shp + 8 + ((ind + 1) * 4),
					      SPLE_LITTLE_ENDIAN);
		      else
			  end = n1;
		      points = end - start;
		      line = sple_alloc_linestring (points);
		      points = 0;
		      for (iv = start; iv < end; iv++)
			{
			    x = sple_import_64 (buf_shp + base +
						(iv * 16), SPLE_LITTLE_ENDIAN);
			    y = sple_import_64 (buf_shp + base +
						(iv * 16) + 8,
						SPLE_LITTLE_ENDIAN);
			    sple_set_point (line, points, x, y);
			    start++;
			    points++;
			}
		      if (!geom)
			  geom = new_geom_line (line);
		      else
			  add_line_to_geom (geom, line);
		  }
	    }
	  if (shape == 5 || shape == 15 || shape == 25)
	    {
		/* shape polygon */
		rd = fread (buf_shp, sizeof (unsigned char), 32, fl_shp);
		if (rd != 32)
		    goto error;
		rd = fread (buf_shp, sizeof (unsigned char),
			    (sz * 2) - 36, fl_shp);
		if (rd != (sz * 2) - 36)
		    goto error;
		n = sple_import_32 (buf_shp, SPLE_LITTLE_ENDIAN);
		n1 = sple_import_32 (buf_shp + 4, SPLE_LITTLE_ENDIAN);
		base = 8 + (n * 4);
		start = 0;
		for (ind = 0; ind < n; ind++)
		  {
		      if (ind < (n - 1))
			  end =
			      sple_import_32 (buf_shp + 8 + ((ind + 1) * 4),
					      SPLE_LITTLE_ENDIAN);
		      else
			  end = n1;
		      points = end - start;
		      line = sple_alloc_linestring (points);
		      points = 0;
		      for (iv = start; iv < end; iv++)
			{
			    x = sple_import_64 (buf_shp + base +
						(iv * 16), SPLE_LITTLE_ENDIAN);
			    y = sple_import_64 (buf_shp + base +
						(iv * 16) + 8,
						SPLE_LITTLE_ENDIAN);
			    sple_set_point (line, points, x, y);
			    start++;
			    points++;
			}
		      if (!geom)
			  geom = new_geom_polyg (line, &polyg);	/* new geometry - new need to allocate a new POLYGON */
		      else
			{
			    if (is_clockwise_ring (line))
			      {
				  /* this one is a POLYGON exterior ring - we need to allocate e new POLYGON */
				  add_polyg_to_geom (geom, line, &polyg);
			      }
			    else
			      {
				  /* adding an interior ring to current POLYGON */
				  add_ring_to_polyg (polyg, line);
			      }
			}
		  }
	    }
	  /* preparing WKB representation for current geometry */
	  sple_blob_to_wkb (geom, &wkb_buf, &wkb_size);
	  sple_free_geometry (geom);
	  geom = NULL;
	  /* reading correspondig DBF entity - attributes */
	  skpos = fseek (fl_dbf, cur_dbf, SEEK_SET);
	  if (skpos < 0)
	      goto error;
	  rd = fread (buf_dbf, sizeof (unsigned char), dbf_reclen, fl_dbf);
	  if (rd != dbf_reclen)
	      goto error;
	  cur_dbf += dbf_reclen;
	  /* preparing INSERT INTO statement */
	  sprintf (sql, "INSERT INTO %s (", table);
	  dbf_field = dbf_list->first;
	  while (dbf_field)
	    {
		/* columns corresponding to some DBF attribute */
		strcat (sql, dbf_field->name);
		strcat (sql, ",");
		dbf_field = dbf_field->next;
	    }
	  strcat (sql, (char *) column);	/* the GEOMETRY column */
	  strcat (sql, ") VALUES (");
	  dbf_field = dbf_list->first;
	  while (dbf_field)
	    {
		/* fetching DBF attribute values */
		memcpy (value, buf_dbf + dbf_field->offset + 1,
			dbf_field->length);
		value[dbf_field->length] = '\0';
		if (dbf_field->type == SPLE_DBF_BOOL)
		  {
		      /* setting a BOOLEAN value */
		      if (*value == 'N' || *value == 'n' ||
			  *value == 'F' || *value == 'f' || *value == '0')
			  strcat (sql, "1,");
		      else
			  strcat (sql, "0,");
		  }
		else if (dbf_field->type == SPLE_DBF_DATE)
		  {
		      /* setting a DATE value */
		      if (strlen (value) != 8)
			  strcat (sql, "'1970-01-01',");
		      else
			{
			    date[0] = '\'';
			    date[1] = value[0];
			    date[2] = value[1];
			    date[3] = value[2];
			    date[4] = value[3];
			    date[5] = '-';
			    date[6] = value[4];
			    date[7] = value[5];
			    date[8] = '-';
			    date[9] = value[6];
			    date[10] = value[7];
			    date[11] = '\'';
			    date[12] = ',';
			    date[13] = '\0';
			    strcat (sql, date);
			}
		  }
		else if (dbf_field->type == SPLE_DBF_NUM)
		  {
		      /* setting a NUMERIC value */
		      for (i = 0; i < strlen (value); i++)
			  if (value[i] == ',')
			      value[i] = '.';	/* sometimes DBF numbers use decimal point as comma */
		      if (dbf_field->decimals)
			{
			    sprintf (fmt, "%%1.%dlf,", dbf_field->decimals);
			    sprintf (dummy, fmt, atof (value));
			}
		      else if (dbf_field->length <= 8)
			  sprintf (dummy, "%ld,", atol (value));
		      else
			  sprintf (dummy, "%1.0lf,", atof (value));
		      strcat (sql, dummy);
		  }
		else
		  {
		      /* setting a TEXT value */
		      clean_sql_string (value);
		      strcat (sql, "'");
		      strcat (sql, value);
		      strcat (sql, "',");
		  }
		dbf_field = dbf_field->next;
	    }
	  /* computing size for real SQL statement, including geometry as WKB */
	  sql_sz =
	      strlen (sql) + strlen ("GeomFromWkb(X'") + (wkb_size * 2) + 64;
	  real_sql = malloc (sql_sz + 1);
	  /* prepares real SQL statement including geometry as WKB */
	  strcpy ((char *) real_sql, sql);
	  strcat ((char *) real_sql, "GeomFromWkb(X'");
	  i_sql = strlen ((char *) real_sql);
	  for (i = 0; i < wkb_size; i++)
	    {
		/* translates WKB in hexadecimal notation */
		sprintf (dummy, "%02X", wkb_buf[i]);
		real_sql[i_sql++] = dummy[0];
		real_sql[i_sql++] = dummy[1];
	    }
	  free (wkb_buf);
	  real_sql[i_sql] = '\0';
	  if (value < 0)
	      strcpy (value, "'))");
	  else
	      sprintf (value, "', %d))", srid);
	  strcat ((char *) real_sql, value);
	  /* finally, we insert a new row in SQLite table */
	  ret = sqlite3_exec (this_db, (char *) real_sql, NULL, 0, &err_sql);
	  free (real_sql);
	  if (ret != SQLITE_OK)
	      goto sql_error;
	  rows++;
      }
    sple_free_dbf_list (dbf_list);
    free (buf_shp);
    free (buf_dbf);
    if (fl_shx)
	fclose (fl_shx);
    if (fl_shp)
	fclose (fl_shp);
    if (fl_dbf)
	fclose (fl_dbf);
    return rows;
  no_file:
/* one of shapefile's files can't be accessed */
    sys_err = strerror (errno);
    if (fl_shx)
	fclose (fl_shx);
    if (fl_shp)
	fclose (fl_shp);
    if (fl_dbf)
	fclose (fl_dbf);
    sprintf (sql, "unable to open '%s' for reading: %s", xpath, sys_err);
    len = strlen (sql);
    *error = malloc (len + 1);
    strcpy (*error, sql);
    return -1;
  error:
/* the shapefile is invalid or corrupted */
    sple_free_dbf_list (dbf_list);
    if (buf_shp)
	free (buf_shp);
    if (buf_dbf)
	free (buf_dbf);
    fclose (fl_shx);
    fclose (fl_shp);
    fclose (fl_dbf);
    sprintf (sql, "invalid or corrupted shapefile '%s'", path);
    len = strlen (sql);
    *error = malloc (len + 1);
    strcpy (*error, sql);
    return -1;
  unsupported:
/* the shapefile has an unrecognized shape type */
    sple_free_dbf_list (dbf_list);
    if (buf_shp)
	free (buf_shp);
    if (buf_dbf)
	free (buf_dbf);
    fclose (fl_shx);
    fclose (fl_shp);
    if (fl_dbf)
	fclose (fl_dbf);
    sprintf (sql, "shapefile '%s' has an unsupported shape type", path);
    len = strlen (sql);
    *error = malloc (len + 1);
    strcpy (*error, sql);
    return -1;
  no_table:
/* SQLite table creation failed */
    sple_free_dbf_list (dbf_list);
    if (buf_shp)
	free (buf_shp);
    if (buf_dbf)
	free (buf_dbf);
    fclose (fl_shx);
    fclose (fl_shp);
    fclose (fl_dbf);
    sprintf (sql, "CREATE TABLE '%s' failed: %s", table, err_sql);
    sqlite3_free (err_sql);
    len = strlen (sql);
    *error = malloc (len + 1);
    strcpy (*error, sql);
    return -1;
  sql_error:
/* some SQL error occurred */
    sple_free_dbf_list (dbf_list);
    if (buf_shp)
	free (buf_shp);
    if (buf_dbf)
	free (buf_dbf);
    fclose (fl_shx);
    fclose (fl_shp);
    fclose (fl_dbf);
    sprintf (sql, "INSERT INTO failed: %s", err_sql);
    sqlite3_free (err_sql);
    len = strlen (sql);
    *error = malloc (len + 1);
    strcpy (*error, sql);
    return -1;
}

SPATIALITE_DECLARE int
sple_dump_shapefile (const unsigned char *path, const unsigned char *table,
		     const unsigned char *column, int *rows, int xshape,
		     char **error)
{
/* trying to dump an SQLite table into an external shapefile  */
    extern sqlite3 *this_db;
    FILE *fl_shx = NULL;
    FILE *fl_shp = NULL;
    FILE *fl_dbf = NULL;
    char xpath[1024];
    sqlite3_stmt *stmt;
    char sql[1024];
    char dummy[1024];
    int ret;
    int len;
    int i;
    int type;
    int entities = 0;
    int n_cols = 0;
    int ie;
    int ib;
    int iv;
    int int_value;
    double double_value;
    const unsigned char *char_value;
    const void *blob_value;
    short dbf_reclen = 0;
    char *dbf_buf = NULL;
    unsigned long dbf_recno = 0;
    unsigned char *buf_shp = NULL;
    int buf_size = 1024;
    long shp_size = 0;
    long shx_size = 0;
    unsigned short dbf_size = 0;
    char *sys_err;
    int row1 = 0;
    double minx = DBL_MAX;
    double miny = DBL_MAX;
    double maxx = DBL_MIN;
    double maxy = DBL_MIN;
    int multiple_entities = 0;
    int tot_ln;
    int tot_v;
    int this_size;
    int ix;
    sple_dbf_export_ptr dbf_export_list = NULL;
    sple_dbf_export_ptr dbf_field;
    sple_geometry_ptr geom;
    sple_point_ptr point;
    sple_linestring_ptr line;
    sple_generic_geometry_ptr pGeo;
    sple_polygon_ptr polyg;
    *rows = 0;
    buf_shp = malloc (buf_size);
/* preparing SQL statement */
    sprintf (sql, "SELECT * FROM %s WHERE GeometryType(%s) = ", table, column);
    if (xshape == SPLE_LINESTRING)
      {
	  strcat (sql, "'LINESTRING' OR GeometryType(");
	  strcat (sql, (char *) column);
	  strcat (sql, ") = 'MULTILINESTRING'");
      }
    else if (xshape == SPLE_POLYGON)
      {
	  strcat (sql, "'POLYGON' OR GeometryType(");
	  strcat (sql, (char *) column);
	  strcat (sql, ") = 'MULTIPOLYGON'");
      }
    else
      {
	  strcat (sql, "'POINT' OR GeometryType(");
	  strcat (sql, (char *) column);
	  strcat (sql, ") = 'MULTIPOINT'");
      }
/* compiling SQL prepared statement */
    ret = sqlite3_prepare_v2 (this_db, sql, strlen (sql), &stmt, NULL);
    if (ret != SQLITE_OK)
	goto sql_error;
    while (1)
      {
	  /* Pass I - scrolling the result set to compute real DBF attributes' sizes and types */
	  ret = sqlite3_step (stmt);
	  if (ret == SQLITE_DONE)
	      break;		/* end of result set */
	  if (ret == SQLITE_ROW)
	    {
		/* processing a result set row */
		row1++;
		if (n_cols == 0)
		  {
		      /* this one is the first row, so we are going to prepare the DBF Fields list */
		      n_cols = sqlite3_column_count (stmt);
		      dbf_export_list =
			  malloc (sizeof (sple_dbf_export) * n_cols);
		      for (i = 0; i < n_cols; i++)
			{
			    /* initializes the DBF export fields */
			    dbf_field = (dbf_export_list + i);
			    strcpy (dummy, sqlite3_column_name (stmt, i));
			    len = strlen (dummy);
			    dbf_field->name = malloc (len + 1);
			    strcpy (dbf_field->name, dummy);
			    dbf_field->sql_type = SQLITE_NULL;
			    dbf_field->max_length = 0;
			    dbf_field->dbf_type = '\0';
			    dbf_field->boolean = 0;
			    dbf_field->date = 0;
			    dbf_field->offset = 0;
			    dbf_field->length = 0;
			    dbf_field->decimals = 0;
			}
		  }
		for (i = 0; i < n_cols; i++)
		  {
		      /* update the DBF export fields analyzing fetched data */
		      dbf_field = (dbf_export_list + i);
		      type = sqlite3_column_type (stmt, i);
		      if (type == SQLITE_BLOB &&
			  strcasecmp ((char *) column,
				      (char *) sqlite3_column_name (stmt,
								    i)) == 0
			  && xshape == SPLE_POINT)
			{
			    /* 
			       / we need to check if there is any MULTIPOINT,
			       / because shapefile handles simple-points shapes and multi-points shapes
			       / in a complete differet way
			     */
			    blob_value = sqlite3_column_blob (stmt, i);
			    len = sqlite3_column_bytes (stmt, i);
			    geom = sple_blob_decode (blob_value, len);
			    if (geom)
			      {
				  if (geom->entities > 1)
				      multiple_entities = 1;
				  sple_free_geometry (geom);
			      }
			}
		      if (type == SQLITE_NULL || type == SQLITE_BLOB)
			  continue;
		      if (type == SQLITE_TEXT)
			{
			    char_value = sqlite3_column_text (stmt, i);
			    len = sqlite3_column_bytes (stmt, i);
			    /* checks if this column may be a DATE one */
			    if (is_date (char_value, len))
			      {
				  if (!(dbf_field->max_length))
				      dbf_field->date = 1;
			      }
			    else
				dbf_field->date = 0;
			    dbf_field->sql_type = SQLITE_TEXT;	/* promoting the column type to TEXT */
			    if (len > dbf_field->max_length)
				dbf_field->max_length = len;
			}
		      else if (type == SQLITE_FLOAT
			       && dbf_field->sql_type != SQLITE_TEXT)
			  dbf_field->sql_type = SQLITE_FLOAT;	/* promoting a numeric column to be DOUBLE */
		      else if (type == SQLITE_INTEGER &&
			       (dbf_field->sql_type == SQLITE_NULL
				|| dbf_field->sql_type == SQLITE_INTEGER))
			{
			    dbf_field->sql_type = SQLITE_INTEGER;	/* promoting a null column to be INTEGER */
			    int_value = sqlite3_column_int (stmt, i);
			    if (int_value == 0 || int_value == 1)
			      {
				  /* checks if this column may be a BOOLEAN one */
				  if (dbf_field->max_length <= 1)
				    {
					dbf_field->max_length = 1;
					dbf_field->boolean = 1;
				    }
			      }
			    else if (int_value <= 9999
				     && dbf_field->max_length <= 4)
			      {
				  dbf_field->max_length = 4;
				  dbf_field->boolean = 0;
			      }
			    else
			      {
				  dbf_field->max_length = 9;
				  dbf_field->boolean = 0;
			      }
			}
		  }
	    }
	  else
	      goto sql_error;
      }
    if (!row1)
	goto empty_result_set;
/* trying to open shapefile files */
    sprintf (xpath, "%s.shx", path);
    fl_shx = fopen (xpath, "wb");
    if (!fl_shx)
	goto no_file;
    sprintf (xpath, "%s.shp", path);
    fl_shp = fopen (xpath, "wb");
    if (!fl_shp)
	goto no_file;
    sprintf (xpath, "%s.dbf", path);
    fl_dbf = fopen (xpath, "wb");
    if (!fl_dbf)
	goto no_file;
/* computing DBF offsets, allocating DBF buffer an alikes */
    dbf_reclen = sple_compute_export_dbf (dbf_export_list, n_cols);
    dbf_buf = malloc (dbf_reclen);
/* writing an empty SHP file header */
    memset (buf_shp, 0, 100);
    fwrite (buf_shp, 1, 100, fl_shp);
    shp_size = 50;		/* note: shapefile [SHP and SHX] counts sizes in WORDS of 16 bits, not in bytes of 8 bits !!!! */
/* writing an empty SHX file header */
    memset (buf_shp, 0, 100);
    fwrite (buf_shp, 1, 100, fl_shx);
    shx_size = 50;
/* writing the DBF file header */
    memset (buf_shp, '\0', 32);
    fwrite (buf_shp, 1, 32, fl_dbf);
    dbf_size = 32;		/* note: DBF counts sizes in bytes */
    for (i = 0; i < n_cols; i++)
      {
	  /* exporting DBF Fields specifications */
	  dbf_field = (dbf_export_list + i);
	  if (dbf_field->sql_type == SQLITE_NULL)	/* this column has not to be exported - containts only NULL or BLOB values */
	      continue;
	  strcpy (dummy, dbf_field->name);
	  dummy[10] = '\0';
	  memset (buf_shp, 0, 32);
	  memcpy (buf_shp, dummy, strlen (dummy));
	  *(buf_shp + 11) = dbf_field->dbf_type;
	  *(buf_shp + 16) = dbf_field->length;
	  *(buf_shp + 17) = dbf_field->decimals;
	  fwrite (buf_shp, 1, 32, fl_dbf);
	  dbf_size += 32;
      }
    fwrite ("\r", 1, 1, fl_dbf);	/* this one is a special DBF delimiter that closes file header */
    dbf_size++;
/* 
/ all initialization tasks completed
/ we can now restart the result set and really export the shapefile
*/
    ret = sqlite3_reset (stmt);
    if (ret != SQLITE_OK)
	goto sql_error;
    while (1)
      {
	  /* Pass II - scrolling the result set to dump data into shapefile */
	  ret = sqlite3_step (stmt);
	  if (ret == SQLITE_DONE)
	      break;		/* end of result set */
	  if (ret == SQLITE_ROW)
	    {
		/* 
		   / processing a result set row 
		   / we start with cleaning the DBF buffer and the GEOMETRY
		 */
		(*rows)++;
		geom = NULL;
		memset (dbf_buf, '\0', dbf_reclen);
		*dbf_buf = ' ';	/* in DBF first byte of each row marks for validity or deletion */
		for (i = 0; i < n_cols; i++)
		  {
		      dbf_field = (dbf_export_list + i);
		      if (dbf_field->sql_type == SQLITE_NULL)
			{
			    if (strcasecmp
				((char *) column,
				 (char *) sqlite3_column_name (stmt, i)) == 0)
			      {
				  /* this one is the internal BLOB encoded GEOMETRY to be exported */
				  blob_value = sqlite3_column_blob (stmt, i);
				  len = sqlite3_column_bytes (stmt, i);
				  geom = sple_blob_decode (blob_value, len);
			      }
			    /* skipping other columns with BLOB or NULL values */
			    continue;
			}
		      switch (dbf_field->dbf_type)
			{
			case 'L':
			    if (sqlite3_column_type (stmt, i) == SQLITE_INTEGER)
			      {
				  /* exporting a BOOLEAN value */
				  int_value = sqlite3_column_int (stmt, i);
				  if (int_value == 1)
				      *(dbf_buf + dbf_field->offset) = 'Y';
				  else
				      *(dbf_buf + dbf_field->offset) = 'N';
			      }
			    break;
			case 'D':
			    if (sqlite3_column_type (stmt, i) == SQLITE_TEXT)
			      {
				  /* exporting a DATE value */
				  char_value = sqlite3_column_text (stmt, i);
				  len = sqlite3_column_bytes (stmt, i);
				  if (len != 10)
				      strcpy (dummy, "00000000");
				  else
				    {
					dummy[0] = char_value[6];
					dummy[1] = char_value[7];
					dummy[2] = char_value[8];
					dummy[3] = char_value[9];
					dummy[4] = char_value[3];
					dummy[5] = char_value[4];
					dummy[6] = char_value[0];
					dummy[7] = char_value[1];
					dummy[8] = '\0';
				    }
				  memcpy (dbf_buf + dbf_field->offset, dummy,
					  8);
			      }
			    break;
			case 'C':
			    if (sqlite3_column_type (stmt, i) == SQLITE_TEXT)
			      {
				  /* exporting a CHAR value */
				  char_value = sqlite3_column_text (stmt, i);
				  len = sqlite3_column_bytes (stmt, i);
				  memset (dbf_buf + dbf_field->offset, ' ',
					  dbf_field->length);
				  if (len < dbf_field->length)
				      memcpy (dbf_buf + dbf_field->offset,
					      char_value, len);
				  else
				      memcpy (dbf_buf + dbf_field->offset,
					      char_value, dbf_field->length);
			      }
			    if (sqlite3_column_type (stmt, i) == SQLITE_FLOAT)
			      {
				  /* exporting a DOUBLE value as CHAR */
				  double_value =
				      sqlite3_column_double (stmt, i);
				  sprintf (dummy, "%1.6lf", double_value);
				  len = strlen (dummy);
				  if (len < dbf_field->length)
				      memcpy (dbf_buf + dbf_field->offset,
					      dummy, len);
				  else
				      memcpy (dbf_buf + dbf_field->offset,
					      dummy, dbf_field->length);
			      }
			    if (sqlite3_column_type (stmt, i) == SQLITE_INTEGER)
			      {
				  /* exporting an INTEGER value as CHAR */
				  int_value = sqlite3_column_int (stmt, i);
				  sprintf (dummy, "%d", int_value);
				  len = strlen (dummy);
				  if (len < dbf_field->length)
				      memcpy (dbf_buf + dbf_field->offset,
					      dummy, len);
				  else
				      memcpy (dbf_buf + dbf_field->offset,
					      dummy, dbf_field->length);
			      }
			    break;
			case 'N':
			    if (sqlite3_column_type (stmt, i) == SQLITE_FLOAT)
			      {
				  /* exporting a DOUBLE value */
				  double_value =
				      sqlite3_column_double (stmt, i);
				  sprintf (dummy, "%1.6lf", double_value);
				  len = strlen (dummy);
				  if (len < dbf_field->length)
				      memcpy (dbf_buf + dbf_field->offset,
					      dummy, len);
				  else
				      memcpy (dbf_buf + dbf_field->offset,
					      dummy, dbf_field->length);
			      }
			    if (sqlite3_column_type (stmt, i) == SQLITE_INTEGER)
			      {
				  /* exporting an INTEGER value */
				  int_value = sqlite3_column_int (stmt, i);
				  sprintf (dummy, "%d", int_value);
				  len = strlen (dummy);
				  if (len < dbf_field->length)
				      memcpy (dbf_buf + dbf_field->offset,
					      dummy, len);
				  else
				      memcpy (dbf_buf + dbf_field->offset,
					      dummy, dbf_field->length);
			      }
			    break;
			};
		  }
		if (geom != NULL)
		  {
		      /* updates the shapefile main MBR */
		      if (geom->minx < minx)
			  minx = geom->minx;
		      if (geom->maxx > maxx)
			  maxx = geom->maxx;
		      if (geom->miny < miny)
			  miny = geom->miny;
		      if (geom->maxy > maxy)
			  maxy = geom->maxy;
		      /*writing into the export files */
		      if (xshape == SPLE_POINT)
			{
			    if (multiple_entities)
			      {
			      }
			    else
			      {
				  /* this one is a plain POINT */
				  pGeo = *(geom->geometries + 0);
				  point = (sple_point_ptr) (pGeo->geometry);
				  /* inserting POINT entity into SHX file */
				  sple_export_32 (buf_shp, shp_size, SPLE_BIG_ENDIAN);	/* exports current SHP file position */
				  sple_export_32 (buf_shp + 4, 10, SPLE_BIG_ENDIAN);	/* exports entitiy size [in 16 bits words !!!] */
				  fwrite (buf_shp, 1, 8, fl_shx);
				  shx_size += 4;
				  /* inserting POINT into SHP file */
				  sple_export_32 (buf_shp, entities, SPLE_BIG_ENDIAN);	/* exports entity ID */
				  sple_export_32 (buf_shp + 4, 10, SPLE_BIG_ENDIAN);	/* exports entity size [in 16 bits words !!!] */
				  sple_export_32 (buf_shp + 8, 1, SPLE_LITTLE_ENDIAN);	/* exports geometry type = POINT */
				  sple_export_64 (buf_shp + 12, point->x, SPLE_LITTLE_ENDIAN);	/* exports X coordinate */
				  sple_export_64 (buf_shp + 20, point->y, SPLE_LITTLE_ENDIAN);	/* exports Y coordinate */
				  fwrite (buf_shp, 1, 28, fl_shp);
				  shp_size += 14;	/* updating current SHP file poisition [in 16 bits words !!!] */
			      }
			}
		      if (xshape == SPLE_LINESTRING)
			{
			    /* this one may be a LINESTRING or a MULTILINESTRING as well */
			    tot_ln = 0;
			    tot_v = 0;
			    for (ie = 0; ie < geom->entities; ie++)
			      {
				  /* computes # lines and total # points */
				  pGeo = *(geom->geometries + ie);
				  line = (sple_linestring_ptr) (pGeo->geometry);
				  tot_v += line->points;
				  tot_ln++;
			      }
			    this_size = 22 + (2 * tot_ln) + (tot_v * 8);	/* size [in 16 bits words !!!] for this SHP entity */
			    if ((this_size * 2) + 1024 > buf_size)
			      {
				  /* current buffer is too small; we need to allocate a bigger one */
				  free (buf_shp);
				  buf_size = (this_size * 2) + 1024;
				  buf_shp = malloc (buf_size);
			      }
			    /* inserting LINESTRING or MULTILINESTRING in SHX file */
			    sple_export_32 (buf_shp, shp_size, SPLE_BIG_ENDIAN);	/* exports current SHP file position */
			    sple_export_32 (buf_shp + 4, this_size, SPLE_BIG_ENDIAN);	/* exports entitiy size [in 16 bits words !!!] */
			    fwrite (buf_shp, 1, 8, fl_shx);
			    shx_size += 4;
			    /* inserting LINESTRING or MULTILINESTRING in SHP file */
			    sple_export_32 (buf_shp, entities, SPLE_BIG_ENDIAN);	/* exports entity ID */
			    sple_export_32 (buf_shp + 4, this_size, SPLE_BIG_ENDIAN);	/* exports entity size [in 16 bits words !!!] */
			    sple_export_32 (buf_shp + 8, 3, SPLE_LITTLE_ENDIAN);	/* exports geometry type = POLYLINE */
			    sple_export_64 (buf_shp + 12, geom->minx, SPLE_LITTLE_ENDIAN);	/* exports the MBR for this geometry */
			    sple_export_64 (buf_shp + 20, geom->miny,
					    SPLE_LITTLE_ENDIAN);
			    sple_export_64 (buf_shp + 28, geom->maxx,
					    SPLE_LITTLE_ENDIAN);
			    sple_export_64 (buf_shp + 36, geom->maxy,
					    SPLE_LITTLE_ENDIAN);
			    sple_export_32 (buf_shp + 44, tot_ln, 'L');	/* exports # lines in this polyline */
			    sple_export_32 (buf_shp + 48, tot_v, 'L');	/* exports total # points */
			    tot_v = 0;	/* resets points counter */
			    ix = 52;	/* sets current buffer offset */
			    for (ie = 0; ie < geom->entities; ie++)
			      {
				  /* exports start point index for each line */
				  pGeo = *(geom->geometries + ie);
				  line = (sple_linestring_ptr) (pGeo->geometry);
				  sple_export_32 (buf_shp + ix, tot_v,
						  SPLE_LITTLE_ENDIAN);
				  tot_v += line->points;
				  ix += 4;
			      }
			    for (ie = 0; ie < geom->entities; ie++)
			      {
				  /* exports points for each line */
				  pGeo = *(geom->geometries + ie);
				  line = (sple_linestring_ptr) (pGeo->geometry);
				  for (iv = 0; iv < line->points; iv++)
				    {
					/* exports a POINT [x,y] */
					sple_export_64 (buf_shp + ix,
							(line->geometry +
							 iv)->x,
							SPLE_LITTLE_ENDIAN);
					ix += 8;
					sple_export_64 (buf_shp + ix,
							(line->geometry +
							 iv)->y,
							SPLE_LITTLE_ENDIAN);
					ix += 8;
				    }
			      }
			    fwrite (buf_shp, 1, ix, fl_shp);
			    shp_size += (ix / 2);	/* updating current SHP file poisition [in 16 bits words !!!] */
			}
		      if (xshape == SPLE_POLYGON)
			{
			    /* this one may be a POLYGON or a MULTIPOLYGON as well */
			    tot_ln = 0;
			    tot_v = 0;
			    for (ie = 0; ie < geom->entities; ie++)
			      {
				  /* computes # rings and total # points */
				  pGeo = *(geom->geometries + ie);
				  polyg = (sple_polygon_ptr) (pGeo->geometry);
				  sane_clockwise (polyg);	/* we must assure that exterior ring is clockwise, and interior rings are anti-clockwise */
				  line = polyg->external;	/* this one is the exterior ring */
				  tot_v += line->points;
				  tot_ln++;
				  for (ib = 0; ib < polyg->borders; ib++)
				    {
					/* that ones are the interior rings */
					line = *(polyg->internals + ib);
					tot_v += line->points;
					tot_ln++;
				    }
			      }
			    this_size = 22 + (2 * tot_ln) + (tot_v * 8);	/* size [in 16 bits words !!!] for this SHP entity */
			    if ((this_size * 2) + 1024 > buf_size)
			      {
				  /* current buffer is too small; we need to allocate a bigger one */
				  free (buf_shp);
				  buf_size = (this_size * 2) + 1024;
				  buf_shp = malloc (buf_size);
			      }
			    /* inserting POLYGON or MULTIPOLYGON in SHX file */
			    sple_export_32 (buf_shp, shp_size, SPLE_BIG_ENDIAN);	/* exports current SHP file position */
			    sple_export_32 (buf_shp + 4, this_size, SPLE_BIG_ENDIAN);	/* exports entitiy size [in 16 bits words !!!] */
			    fwrite (buf_shp, 1, 8, fl_shx);
			    shx_size += 4;
			    /* inserting POLYGON or MULTIPOLYGON in SHP file */
			    sple_export_32 (buf_shp, entities, SPLE_BIG_ENDIAN);	/* exports entity ID */
			    sple_export_32 (buf_shp + 4, this_size, SPLE_BIG_ENDIAN);	/* exports entity size [in 16 bits words !!!] */
			    sple_export_32 (buf_shp + 8, 5, SPLE_LITTLE_ENDIAN);	/* exports geometry type = POLYGON */
			    sple_export_64 (buf_shp + 12, geom->minx, SPLE_LITTLE_ENDIAN);	/* exports the MBR for this geometry */
			    sple_export_64 (buf_shp + 20, geom->miny,
					    SPLE_LITTLE_ENDIAN);
			    sple_export_64 (buf_shp + 28, geom->maxx,
					    SPLE_LITTLE_ENDIAN);
			    sple_export_64 (buf_shp + 36, geom->maxy,
					    SPLE_LITTLE_ENDIAN);
			    sple_export_32 (buf_shp + 44, tot_ln, 'L');	/* exports # rings in this polygon */
			    sple_export_32 (buf_shp + 48, tot_v, 'L');	/* exports total # points */
			    tot_v = 0;	/* resets points counter */
			    ix = 52;	/* sets current buffer offset */
			    for (ie = 0; ie < geom->entities; ie++)
			      {
				  /* exports start point index for each line */
				  pGeo = *(geom->geometries + ie);
				  polyg = (sple_polygon_ptr) (pGeo->geometry);
				  line = polyg->external;	/* this one is the exterior ring */
				  sple_export_32 (buf_shp + ix, tot_v,
						  SPLE_LITTLE_ENDIAN);
				  tot_v += line->points;
				  ix += 4;
				  for (ib = 0; ib < polyg->borders; ib++)
				    {
					/* that ones are the interior rings */
					line = *(polyg->internals + ib);
					sple_export_32 (buf_shp + ix, tot_v,
							SPLE_LITTLE_ENDIAN);
					tot_v += line->points;
					ix += 4;
				    }
			      }
			    for (ie = 0; ie < geom->entities; ie++)
			      {
				  /* exports points for each ring */
				  pGeo = *(geom->geometries + ie);
				  polyg = (sple_polygon_ptr) (pGeo->geometry);
				  line = polyg->external;	/* this one is the exterior ring */
				  for (iv = 0; iv < line->points; iv++)
				    {
					/* exports a POINT [x,y] - exterior ring */
					sple_export_64 (buf_shp + ix,
							(line->geometry +
							 iv)->x,
							SPLE_LITTLE_ENDIAN);
					ix += 8;
					sple_export_64 (buf_shp + ix,
							(line->geometry +
							 iv)->y,
							SPLE_LITTLE_ENDIAN);
					ix += 8;
				    }
				  for (ib = 0; ib < polyg->borders; ib++)
				    {
					/* that ones are the interior rings */
					line = *(polyg->internals + ib);
					for (iv = 0; iv < line->points; iv++)
					  {
					      /* exports a POINT [x,y] - interior ring */
					      sple_export_64 (buf_shp + ix,
							      (line->geometry +
							       iv)->x,
							      SPLE_LITTLE_ENDIAN);
					      ix += 8;
					      sple_export_64 (buf_shp + ix,
							      (line->geometry +
							       iv)->y,
							      SPLE_LITTLE_ENDIAN);
					      ix += 8;
					  }
				    }
			      }
			    fwrite (buf_shp, 1, ix, fl_shp);
			    shp_size += (ix / 2);	/* updating current SHP file poisition [in 16 bits words !!!] */
			}
		      /* inserting entity in DBF file */
		      fwrite (dbf_buf, 1, dbf_reclen, fl_dbf);
		      dbf_recno++;
		      entities++;
		      /* destroying geometry */
		      sple_free_geometry (geom);
		  }
	    }
	  else
	      goto sql_error;
      }
/* writing the SHP file header */
    fseek (fl_shp, 0, SEEK_SET);	/* repositioning at SHP file start */
    sple_export_32 (buf_shp, 9994, SPLE_BIG_ENDIAN);	/* SHP magic number */
    sple_export_32 (buf_shp + 4, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 8, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 12, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 16, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 20, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 24, shp_size, SPLE_BIG_ENDIAN);	/* SHP file size - measured in 16 bits words !!! */
    sple_export_32 (buf_shp + 28, 1000, SPLE_LITTLE_ENDIAN);	/* version */
    if (xshape == SPLE_POINT && !multiple_entities)
	sple_export_32 (buf_shp + 32, 1, SPLE_LITTLE_ENDIAN);	/* ESRI shape POINT */
    if (xshape == SPLE_LINESTRING)
	sple_export_32 (buf_shp + 32, 3, SPLE_LITTLE_ENDIAN);	/* ESRI shape POLYLINE */
    if (xshape == SPLE_POLYGON)
	sple_export_32 (buf_shp + 32, 5, SPLE_LITTLE_ENDIAN);	/* ESRI shape POLYGON */
    sple_export_64 (buf_shp + 36, minx, SPLE_LITTLE_ENDIAN);	/* the MBR for the whole shapefile */
    sple_export_64 (buf_shp + 44, miny, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 52, maxx, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 60, maxy, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 68, 0.0, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 76, 0.0, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 84, 0.0, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 92, 0.0, SPLE_LITTLE_ENDIAN);
    fwrite (buf_shp, 1, 100, fl_shp);
/* writing the SHX file header */
    fseek (fl_shx, 0, SEEK_SET);	/* repositioning at SHX file start */
    sple_export_32 (buf_shp, 9994, SPLE_BIG_ENDIAN);	/* SHP magic number */
    sple_export_32 (buf_shp + 4, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 8, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 12, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 16, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 20, 0, SPLE_BIG_ENDIAN);
    sple_export_32 (buf_shp + 24, shx_size, SPLE_BIG_ENDIAN);	/* SHXfile size - measured in 16 bits words !!! */
    sple_export_32 (buf_shp + 28, 1000, SPLE_LITTLE_ENDIAN);	/* version */
    if (xshape == SPLE_POINT && !multiple_entities)
	sple_export_32 (buf_shp + 32, 1, SPLE_LITTLE_ENDIAN);	/* ESRI shape POINT */
    if (xshape == SPLE_LINESTRING)
	sple_export_32 (buf_shp + 32, 3, SPLE_LITTLE_ENDIAN);	/* ESRI shape POLYLINE */
    if (xshape == SPLE_POLYGON)
	sple_export_32 (buf_shp + 32, 5, SPLE_LITTLE_ENDIAN);	/* ESRI shape POLYGON */
    sple_export_64 (buf_shp + 36, minx, SPLE_LITTLE_ENDIAN);	/* the MBR for the whole shapefile */
    sple_export_64 (buf_shp + 44, miny, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 52, maxx, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 60, maxy, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 68, 0.0, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 76, 0.0, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 84, 0.0, SPLE_LITTLE_ENDIAN);
    sple_export_64 (buf_shp + 92, 0.0, SPLE_LITTLE_ENDIAN);
    fwrite (buf_shp, 1, 100, fl_shx);
/* writing the DBF file header */
    *buf_shp = 0x1a;		/* DBF - this is theEOF marker */
    fwrite (buf_shp, 1, 1, fl_dbf);
    fseek (fl_dbf, 0, SEEK_SET);	/* repositioning at DBF file start */
    memset (buf_shp, '\0', 32);
    *buf_shp = 0x03;		/* DBF magic number */
    *(buf_shp + 1) = 1;		/* this is supposed to be the last update date [Year, Month, Day], but we ignore it at all */
    *(buf_shp + 2) = 1;
    *(buf_shp + 3) = 1;
    sple_export_32 (buf_shp + 4, dbf_recno, SPLE_LITTLE_ENDIAN);	/* exports # records in this DBF */
    sple_export_16 (buf_shp + 8, dbf_size, SPLE_LITTLE_ENDIAN);	/* exports the file header size */
    sple_export_16 (buf_shp + 10, dbf_reclen, SPLE_LITTLE_ENDIAN);	/* exports the record length */
    fwrite (buf_shp, 1, 32, fl_dbf);
    ret = sqlite3_finalize (stmt);
    if (ret != SQLITE_OK)
	goto sql_error;
    sple_free_export_dbf (dbf_export_list, n_cols);
    free (dbf_buf);
    free (buf_shp);
    fclose (fl_shx);
    fclose (fl_shp);
    fclose (fl_dbf);
    return entities;
  sql_error:
/* some SQL error occurred */
    sqlite3_finalize (stmt);
    sple_free_export_dbf (dbf_export_list, n_cols);
    if (dbf_buf)
	free (dbf_buf);
    if (buf_shp)
	free (buf_shp);
    if (fl_shx)
	fclose (fl_shx);
    if (fl_shp)
	fclose (fl_shp);
    if (fl_dbf)
	fclose (fl_dbf);
    sprintf (sql, "SELECT failed: %s", sqlite3_errmsg (this_db));
    len = strlen (sql);
    *error = malloc (len + 1);
    strcpy (*error, sql);
    return -1;
  no_file:
/* one of shapefile's files can't be created/opened */
    sys_err = strerror (errno);
    sqlite3_finalize (stmt);
    sple_free_export_dbf (dbf_export_list, n_cols);
    if (dbf_buf)
	free (dbf_buf);
    if (buf_shp)
	free (buf_shp);
    if (fl_shx)
	fclose (fl_shx);
    if (fl_shp)
	fclose (fl_shp);
    if (fl_dbf)
	fclose (fl_dbf);
    sprintf (sql, "unable to open '%s' for writing: %s", xpath, sys_err);
    len = strlen (sql);
    *error = malloc (len + 1);
    strcpy (*error, sql);
    return -1;
  empty_result_set:
/* the result set is empty - nothing to do */
    sqlite3_finalize (stmt);
    sple_free_export_dbf (dbf_export_list, n_cols);
    if (dbf_buf)
	free (dbf_buf);
    if (buf_shp)
	free (buf_shp);
    if (fl_shx)
	fclose (fl_shx);
    if (fl_shp)
	fclose (fl_shp);
    if (fl_dbf)
	fclose (fl_dbf);
    strcpy (sql,
	    "The SQL SELECT returned an empty result set ... there is nothing to export ...");
    len = strlen (sql);
    *error = malloc (len + 1);
    strcpy (*error, sql);
    return -1;
}

static int
create_table (const unsigned char *table, const unsigned char *column,
	      sple_dbf_list_ptr list, char **msg)
{
/* trying to create the SQLite table */
    extern sqlite3 *this_db;
    char sql[4096];
    int ret;
    sple_dbf_field_ptr field;
    sprintf (sql, "CREATE TABLE ");
    strcat (sql, (char *) table);
    strcat (sql, " (PK_UID INTEGER PRIMARY KEY AUTOINCREMENT,");
    field = list->first;
    while (field)
      {
	  strcat (sql, field->name);
	  strcat (sql, " ");
	  switch (field->type)
	    {
	    case SPLE_DBF_CHAR:
		strcat (sql, "TEXT");
		break;
	    case SPLE_DBF_NUM:
		if (field->decimals)
		    strcat (sql, "DOUBLE");
		else
		  {
		      if (field->length <= 8)
			  strcat (sql, "INTEGER");
		      else
			  strcat (sql, "DOUBLE");
		  }
		break;
	    case SPLE_DBF_BOOL:
		strcat (sql, "BOOLEAN");
		break;
	    case SPLE_DBF_DATE:
		strcat (sql, "DATE");
		break;
	    };
	  strcat (sql, " NOT NULL,");
	  field = field->next;
      }
    strcat (sql, (char *) column);
    strcat (sql, " BLOB NOT NULL)");
    ret = sqlite3_exec (this_db, sql, NULL, 0, msg);
    if (ret != SQLITE_OK)
	return 0;
    return 1;
}

SPATIALITE_DECLARE sple_dbf_field_ptr
sple_alloc_dbf_field (char *name, unsigned char type,
		      int offset, unsigned char length, unsigned char decimals)
{
/* allocates and initializes a DBF Field definition */
    sple_dbf_field_ptr p = malloc (sizeof (sple_dbf_field));
    int len = strlen (name);
    p->name = malloc (len + 1);
    strcpy (p->name, name);
    p->type = type;
    p->offset = offset;
    p->length = length;
    p->decimals = decimals;
    p->next = NULL;
    return p;
}

SPATIALITE_DECLARE void
sple_free_dbf_field (sple_dbf_field_ptr p)
{
/* frees all memory allocations for this DBF Field definition */
    if (!p)
	return;
    if (p->name)
	free (p->name);
    free (p);
}

SPATIALITE_DECLARE sple_dbf_list_ptr
sple_alloc_dbf_list (char *primarykey, char *geocolumn)
{
/* allocates and initializes the DBF Fields list */
    sple_dbf_list_ptr list = malloc (sizeof (sple_dbf_list));
    list->primary_key_name = primarykey;
    list->geometry_name = geocolumn;
    list->first = NULL;
    list->last = NULL;
    list->seed = 1;
    return list;
}

SPATIALITE_DECLARE void
sple_free_dbf_list (sple_dbf_list_ptr list)
{
/* frees all memory allocations related to DBF Fields list */
    sple_dbf_field_ptr p;
    sple_dbf_field_ptr pn;
    if (!list)
	return;
    p = list->first;
    while (p)
      {
	  pn = p->next;
	  sple_free_dbf_field (p);
	  p = pn;
      }
    free (p);
}

SPATIALITE_DECLARE void
sple_add_dbf_field (sple_dbf_list_ptr list, char *name, unsigned char type,
		    int offset, unsigned char length, unsigned char decimals)
{
/* inserts a Field in the DBF Fields list */
    char alias[64];
    if (!list)
	return;
    strcpy (alias, name);
    while (1)
      {
	  if (is_duplicate_column (list, alias))
	    {
		/* 
		   / the original DBF attribute name causes a non-unique column name; 
		   / so it will try an alias name
		 */
		sprintf (alias, "%s_X%d", name, (list->seed)++);
		continue;
	    }
	  if (is_reserved_sql_name (alias))
	    {
		/* 
		   / the original DBF attribute name is an SQL reserver keyword: 
		   / so it will try an alias name
		 */
		sprintf (alias, "%s_X%d", name, (list->seed)++);
		continue;
	    }
	  if (is_reserved_sqlite_name (alias))
	    {
		/* 
		   / the original DBF attribute name is an SQLite reserver keyword: 
		   / so it will try an alias name
		 */
		sprintf (alias, "%s_X%d", name, (list->seed)++);
		continue;
	    }
	  /* the column name is valid */
	  break;
      }
    sple_dbf_field_ptr p =
	sple_alloc_dbf_field (alias, type, offset, length, decimals);
    if (!(list->first))
	list->first = p;
    if (list->last)
	list->last->next = p;
    list->last = p;
}

static int
is_duplicate_column (sple_dbf_list_ptr list, char *name)
{
/* checks if column-name is non-unique */
    sple_dbf_field_ptr p = list->first;
    while (p)
      {
	  if (strcasecmp (p->name, name) == 0)
	      return 1;
	  p = p->next;
      }
    if (strcasecmp (list->primary_key_name, name) == 0)
	return 1;
    if (strcasecmp (list->geometry_name, name) == 0)
	return 1;
    return 0;
}

static int
is_reserved_sqlite_name (char *name)
{
/* checks if column-name is an SQLite reserved keyword */
    char *reserved[] = {
	"ALL",
	"ALTER",
	"AND",
	"AS",
	"AUTOINCREMENT",
	"BETWEEN",
	"BY",
	"CASE",
	"CHECK",
	"COLLATE",
	"COMMIT",
	"CONSTRAINT",
	"CREATE",
	"CROSS",
	"DEFAULT",
	"DEFERRABLE",
	"DELETE",
	"DISTINCT",
	"DROP",
	"ELSE",
	"ESCAPE",
	"EXCEPT",
	"FOREIGN",
	"FROM",
	"FULL",
	"GLOB",
	"GROUP",
	"HAVING",
	"IN",
	"INDEX",
	"INNER",
	"INSERT",
	"INTERSECT",
	"INTO",
	"IS",
	"ISNULL",
	"JOIN",
	"LEFT",
	"LIKE",
	"LIMIT",
	"NATURAL",
	"NOT",
	"NOTNULL",
	"NULL",
	"ON",
	"OR",
	"ORDER",
	"OUTER",
	"PRIMARY",
	"REFERENCES",
	"RIGHT",
	"ROLLBACK",
	"SELECT",
	"SET",
	"TABLE",
	"THEN",
	"TO",
	"TRANSACTION",
	"UNION",
	"UNIQUE",
	"UPDATE",
	"USING",
	"VALUES",
	"WHEN",
	"WHERE",
	NULL
    };
    char **pw = reserved;
    while (*pw != NULL)
      {
	  if (strcasecmp (name, *pw) == 0)
	      return 1;
	  pw++;
      }
    return 0;
}

static int
is_reserved_sql_name (char *name)
{
/* checks if column-name is an SQL reserved keyword */
    char *reserved[] = {
	"ABSOLUTE",
	"ACTION",
	"ADD",
	"AFTER",
	"ALL",
	"ALLOCATE",
	"ALTER",
	"AND",
	"ANY",
	"ARE",
	"ARRAY",
	"AS",
	"ASC",
	"ASENSITIVE",
	"ASSERTION",
	"ASYMMETRIC",
	"AT",
	"ATOMIC",
	"AUTHORIZATION",
	"AVG",
	"BEFORE",
	"BEGIN",
	"BETWEEN",
	"BIGINT",
	"BINARY",
	"BIT",
	"BIT_LENGTH",
	"BLOB",
	"BOOLEAN",
	"BOTH",
	"BREADTH",
	"BY",
	"CALL",
	"CALLED",
	"CASCADE",
	"CASCADED",
	"CASE",
	"CAST",
	"CATALOG",
	"CHAR",
	"CHARACTER",
	"CHARACTER_LENGTH",
	"CHAR_LENGTH",
	"CHECK",
	"CLOB",
	"CLOSE",
	"COALESCE",
	"COLLATE",
	"COLLATION",
	"COLUMN",
	"COMMIT",
	"CONDITION",
	"CONNECT",
	"CONNECTION",
	"CONSTRAINT",
	"CONSTRAINTS",
	"CONSTRUCTOR",
	"CONTAINS",
	"CONTINUE",
	"CONVERT",
	"CORRESPONDING",
	"COUNT",
	"CREATE",
	"CROSS",
	"CUBE",
	"CURRENT",
	"CURRENT_DATE",
	"CURRENT_DEFAULT_TRANSFORM_GROUP",
	"CURRENT_PATH",
	"CURRENT_ROLE",
	"CURRENT_TIME",
	"CURRENT_TIMESTAMP",
	"CURRENT_TRANSFORM_GROUP_FOR_TYPE",
	"CURRENT_USER",
	"CURSOR",
	"CYCLE",
	"DATA",
	"DATE",
	"DAY",
	"DEALLOCATE",
	"DEC",
	"DECIMAL",
	"DECLARE",
	"DEFAULT",
	"DEFERRABLE",
	"DEFERRED",
	"DELETE",
	"DEPTH",
	"DEREF",
	"DESC",
	"DESCRIBE",
	"DESCRIPTOR",
	"DETERMINISTIC",
	"DIAGNOSTICS",
	"DISCONNECT",
	"DISTINCT",
	"DO",
	"DOMAIN",
	"DOUBLE",
	"DROP",
	"DYNAMIC",
	"EACH",
	"ELEMENT",
	"ELSE",
	"ELSEIF",
	"END",
	"EQUALS",
	"ESCAPE",
	"EXCEPT",
	"EXCEPTION",
	"EXEC",
	"EXECUTE",
	"EXISTS",
	"EXIT",
	"external",
	"EXTRACT",
	"FALSE",
	"FETCH",
	"FILTER",
	"FIRST",
	"FLOAT",
	"FOR",
	"FOREIGN",
	"FOUND",
	"FREE",
	"FROM",
	"FULL",
	"FUNCTION",
	"GENERAL",
	"GET",
	"GLOBAL",
	"GO",
	"GOTO",
	"GRANT",
	"GROUP",
	"GROUPING",
	"HANDLER",
	"HAVING",
	"HOLD",
	"HOUR",
	"IDENTITY",
	"IF",
	"IMMEDIATE",
	"IN",
	"INDICATOR",
	"INITIALLY",
	"INNER",
	"INOUT",
	"INPUT",
	"INSENSITIVE",
	"INSERT",
	"INT",
	"INTEGER",
	"INTERSECT",
	"INTERVAL",
	"INTO",
	"IS",
	"ISOLATION",
	"ITERATE",
	"JOIN",
	"KEY",
	"LANGUAGE",
	"LARGE",
	"LAST",
	"LATERAL",
	"LEADING",
	"LEAVE",
	"LEFT",
	"LEVEL",
	"LIKE",
	"LOCAL",
	"LOCALTIME",
	"LOCALTIMESTAMP",
	"LOCATOR",
	"LOOP",
	"LOWER",
	"MAP",
	"MATCH",
	"MAX",
	"MEMBER",
	"MERGE",
	"METHOD",
	"MIN",
	"MINUTE",
	"MODIFIES",
	"MODULE",
	"MONTH",
	"MULTISET",
	"NAMES",
	"NATIONAL",
	"NATURAL",
	"NCHAR",
	"NCLOB",
	"NEW",
	"NEXT",
	"NO",
	"NONE",
	"NOT",
	"NULL",
	"NULLIF",
	"NUMERIC",
	"OBJECT",
	"OCTET_LENGTH",
	"OF",
	"OLD",
	"ON",
	"ONLY",
	"OPEN",
	"OPTION",
	"OR",
	"ORDER",
	"ORDINALITY",
	"OUT",
	"OUTER",
	"OUTPUT",
	"OVER",
	"OVERLAPS",
	"PAD",
	"PARAMETER",
	"PARTIAL",
	"PARTITION",
	"PATH",
	"POSITION",
	"PRECISION",
	"PREPARE",
	"PRESERVE",
	"PRIMARY",
	"PRIOR",
	"PRIVILEGES",
	"PROCEDURE",
	"PUBLIC",
	"RANGE",
	"READ",
	"READS",
	"REAL",
	"RECURSIVE",
	"REF",
	"REFERENCES",
	"REFERENCING",
	"RELATIVE",
	"RELEASE",
	"REPEAT",
	"RESIGNAL",
	"RESTRICT",
	"RESULT",
	"RETURN",
	"RETURNS",
	"REVOKE",
	"RIGHT",
	"ROLE",
	"ROLLBACK",
	"ROLLUP",
	"ROUTINE",
	"ROW",
	"ROWS",
	"SAVEPOINT",
	"SCHEMA",
	"SCOPE",
	"SCROLL",
	"SEARCH",
	"SECOND",
	"SECTION",
	"SELECT",
	"SENSITIVE",
	"SESSION",
	"SESSION_USER",
	"SET",
	"SETS",
	"SIGNAL",
	"SIMILAR",
	"SIZE",
	"SMALLINT",
	"SOME",
	"SPACE",
	"SPECIFIC",
	"SPECIFICTYPE",
	"SQL",
	"SQLCODE",
	"SQLERROR",
	"SQLEXCEPTION",
	"SQLSTATE",
	"SQLWARNING",
	"START",
	"STATE",
	"STATIC",
	"SUBMULTISET",
	"SUBSTRING",
	"SUM",
	"SYMMETRIC",
	"SYSTEM",
	"SYSTEM_USER",
	"TABLE",
	"TABLESAMPLE",
	"TEMPORARY",
	"THEN",
	"TIME",
	"TIMESTAMP",
	"TIMEZONE_HOUR",
	"TIMEZONE_MINUTE",
	"TO",
	"TRAILING",
	"TRANSACTION",
	"TRANSLATE",
	"TRANSLATION",
	"TREAT",
	"TRIGGER",
	"TRIM",
	"TRUE",
	"UNDER",
	"UNDO",
	"UNION",
	"UNIQUE",
	"UNKNOWN",
	"UNNEST",
	"UNTIL",
	"UPDATE",
	"UPPER",
	"USAGE",
	"USER",
	"USING",
	"VALUE",
	"VALUES",
	"VARCHAR",
	"VARYING",
	"VIEW",
	"WHEN",
	"WHENEVER",
	"WHERE",
	"WHILE",
	"WINDOW",
	"WITH",
	"WITHIN",
	"WITHOUT",
	"WORK",
	"WRITE",
	"YEAR",
	"ZONE",
	NULL
    };
    char **pw = reserved;
    while (*pw != NULL)
      {
	  if (strcasecmp (name, *pw) == 0)
	      return 1;
	  pw++;
      }
    return 0;
}

static sple_geometry_ptr
new_geom_point (sple_point_ptr point)
{
/*
/ allocates and initializes a new geometry - POINT */
    sple_geometry_ptr geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = 1;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(geom->geometries) =
	sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
    return geom;
}

static sple_geometry_ptr
new_geom_line (sple_linestring_ptr line)
{
/*
/ allocates and initializes a new geometry - LINESTRING */
    sple_geometry_ptr geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = 1;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(geom->geometries) =
	sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) line);
    return geom;
}

static void
add_line_to_geom (sple_geometry_ptr geom, sple_linestring_ptr line)
{
/* inserts an other line in this geometry - MULTILINESTRING */
    sple_generic_geometry_ptr *old_geometries = geom->geometries;
    int ie;
    geom->geometries =
	malloc (sizeof (sple_generic_geometry_ptr) * (geom->entities + 1));
    for (ie = 0; ie < geom->entities; ie++)
	*(geom->geometries + ie) = *(old_geometries + ie);
    *(geom->geometries + geom->entities) =
	sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) line);
    (geom->entities)++;
    free (old_geometries);
}

static sple_geometry_ptr
new_geom_polyg (sple_linestring_ptr ring, sple_polygon_ptr * pg)
{
/*
/ allocates and initializes a new geometry - POLYGON */
    sple_polygon_ptr polyg;
    sple_geometry_ptr geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = 1;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    polyg = malloc (sizeof (sple_polygon));
    *pg = polyg;
    polyg->external = ring;
    polyg->borders = 0;
    polyg->internals = NULL;
    *(geom->geometries) =
	sple_alloc_generic_geometry (SPLE_POLYGON, (void *) polyg);
    return geom;
}

static void
add_polyg_to_geom (sple_geometry_ptr geom, sple_linestring_ptr ring,
		   sple_polygon_ptr * pg)
{
/* inserts an other line in this geometry - MULTIPOLYGON */
    sple_polygon_ptr polyg;
    sple_generic_geometry_ptr *old_geometries = geom->geometries;
    int ie;
    geom->geometries =
	malloc (sizeof (sple_generic_geometry_ptr) * (geom->entities + 1));
    for (ie = 0; ie < geom->entities; ie++)
	*(geom->geometries + ie) = *(old_geometries + ie);
    polyg = malloc (sizeof (sple_polygon));
    *pg = polyg;
    polyg->external = ring;
    polyg->borders = 0;
    polyg->internals = NULL;
    *(geom->geometries + geom->entities) =
	sple_alloc_generic_geometry (SPLE_POLYGON, (void *) polyg);
    (geom->entities)++;
    free (old_geometries);
}

static void
add_ring_to_polyg (sple_polygon_ptr polyg, sple_linestring_ptr ring)
{
/* adds an interior ring to last POLYGON of this geometry */
    int ib;
    sple_linestring_ptr *old_internals = NULL;
    if (!(polyg->borders))
      {
	  /* this one is the first interior ring */
	  polyg->internals = malloc (sizeof (sple_linestring_ptr));
	  *(polyg->internals) = ring;
	  polyg->borders = 1;
      }
    else
      {
	  /* adding another interior ring */
	  old_internals = polyg->internals;
	  polyg->internals =
	      malloc (sizeof (sple_linestring_ptr) * (polyg->borders + 1));
	  for (ib = 0; ib < polyg->borders; ib++)
	      *(polyg->internals + ib) = *(old_internals + ib);
	  *(polyg->internals + polyg->borders) = ring;
	  (polyg->borders)++;
	  free (old_internals);
      }
}

static int
is_clockwise_ring (sple_linestring_ptr ring)
{
/* checks if this ring is clockwise or anti-clockwise */
    int iv;
    int ix;
    double xx;
    double yy;
    double x;
    double y;
    double area = 0.0;
    for (iv = 0; iv < ring->points; iv++)
      {
	  xx = (ring->geometry + iv)->x;
	  yy = (ring->geometry + iv)->y;
	  ix = (iv + 1) % ring->points;
	  x = (ring->geometry + ix)->x;
	  y = (ring->geometry + ix)->y;
	  area += ((xx * y) - (x * yy));
      }
    area /= 2.0;
    if (area >= 0.0)
	return 0;
    else
	return 1;
}

static void
clean_sql_string (char *value)
{
/*
/ returns a well formatted TEXT value for SQL
/ 1] strips trailing spaces
/ 2] masks any ' inside the string, appending another '
*/
    char new_value[1024];
    char *p;
    int len;
    int i;
    len = strlen (value);
    for (i = (len - 1); i >= 0; i--)
      {
	  /* stripping trailing spaces */
	  if (value[i] == ' ')
	      value[i] = '\0';
	  else
	      break;
      }
    p = new_value;
    for (i = 0; i < len; i++)
      {
	  if (value[i] == '\'')
	      *(p++) = '\'';
	  *(p++) = value[i];
      }
    *p = '\0';
    strcpy (value, new_value);
}

SPATIALITE_DECLARE int
sple_compute_export_dbf (sple_dbf_export_ptr list, int elements)
{
/* 
/ compute real DBF fields types, offsets ans sizes
/
/ returns the DBF record length
*/
    int i;
    int len = 1;		/* an extra byte is needed because in DBF rows first byte is a marker for deletion */
    sple_dbf_export_ptr dbf_field;
    for (i = 0; i < elements; i++)
      {
	  dbf_field = (list + i);
	  if (dbf_field->sql_type == SQLITE_NULL)
	      continue;		/* this attribute has not to be exported; we can ignore it */
	  if (dbf_field->sql_type == SQLITE_TEXT)
	    {
		if (dbf_field->date)
		  {
		      /* this DBF attribute has a DATE type */
		      dbf_field->dbf_type = 'D';
		      dbf_field->length = 8;
		  }
		else
		  {
		      /* normal CHARACTER type */
		      dbf_field->dbf_type = 'C';
		      if (dbf_field->max_length > 255)
			  dbf_field->length = 255;
		      else
			  dbf_field->length = dbf_field->max_length;
		  }
		dbf_field->offset = len;
		len += dbf_field->length;
		dbf_field->decimals = 0;
	    }
	  if (dbf_field->sql_type == SQLITE_FLOAT)
	    {
		dbf_field->dbf_type = 'N';
		dbf_field->length = 24;
		dbf_field->offset = len;
		len += dbf_field->length;
		dbf_field->decimals = 6;
	    }
	  if (dbf_field->sql_type == SQLITE_INTEGER)
	    {
		dbf_field->dbf_type = 'N';
		if (dbf_field->max_length == 1 && dbf_field->boolean)
		  {
		      /* this DBF attribute has a BOOLEAN type */
		      dbf_field->dbf_type = 'L';
		      dbf_field->length = 1;
		  }
		else if (dbf_field->max_length <= 4)
		    dbf_field->length = 5;
		else
		    dbf_field->length = 9;
		dbf_field->offset = len;
		len += dbf_field->length;
		dbf_field->decimals = 0;
	    }
      }
    return len;
}


SPATIALITE_DECLARE void
sple_free_export_dbf (sple_dbf_export_ptr list, int elements)
{
/* frees all memory allocations related to DBF attributes export list */
    int i;
    sple_dbf_export_ptr dbf_field;
    for (i = 0; i < elements; i++)
      {
	  dbf_field = (list + i);
	  if (dbf_field->name)
	      free (dbf_field->name);
      }
    free (list);
}

static int
is_date (const unsigned char *value, int len)
{
/* 
/ checks if this value may be a DATE 
/
/ return 1 if TRUE
/ or 0 if FALSE
*/
    if (len != 10)
	return 0;
    if (value[0] < '0' || value[0] > '9')
	return 0;
    if (value[1] < '0' || value[1] > '9')
	return 0;
    if (value[2] < '0' || value[2] > '9')
	return 0;
    if (value[3] < '0' || value[3] > '9')
	return 0;
    if (value[4] != '-')
	return 0;
    if (value[5] < '0' || value[5] > '9')
	return 0;
    if (value[6] < '0' || value[6] > '9')
	return 0;
    if (value[7] != '-')
	return 0;
    if (value[8] < '0' || value[8] > '9')
	return 0;
    if (value[9] < '0' || value[9] > '9')
	return 0;
    return 1;
}

static void
sane_clockwise (sple_polygon_ptr polyg)
{
/*
/ when exporting POLYGONs to SHAPEFILE, we must guarantee that:
/ - all EXTERIOR RING must be clockwise
/ - all INTERIOR RING must be anti-clockwise
/
/ this function checks for the above conditions,
/ and if needed inverts the rings
*/
    int ib;
    int iv;
    int iv2;
    sple_linestring_ptr new_line;
    sple_linestring_ptr line = polyg->external;
    if (!is_clockwise_ring (line))
      {
	  /* exterior ring needs inversion */
	  new_line = sple_alloc_linestring (line->points);
	  iv2 = 0;
	  for (iv = line->points - 1; iv >= 0; iv--)
	      sple_set_point (new_line, iv2++, (line->geometry + iv)->x,
			      (line->geometry + iv)->y);
	  polyg->external = new_line;
	  sple_free_linestring (line);
      }
    for (ib = 0; ib < polyg->borders; ib++)
      {
	  line = *(polyg->internals + ib);
	  if (is_clockwise_ring (line))
	    {
		/* interior ring needs inversion */
		new_line = sple_alloc_linestring (line->points);
		iv2 = 0;
		for (iv = line->points - 1; iv >= 0; iv--)
		    sple_set_point (new_line, iv2++, (line->geometry + iv)->x,
				    (line->geometry + iv)->y);
		*(polyg->internals + ib) = new_line;
		sple_free_linestring (line);
	    }
      }
}
