/*
/
/ sple_wkt.c - SpatiaLite - spatial extension for SQLite - functions for WKT parsing
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2007  Alessandro Furieri
/
/ This library is free software; you can redistribute it and/or
/ modify it under the terms of the GNU Lesser General Public
/ License as published by the Free Software Foundation; either
/ version 2.1 of the License, or (at your option) any later version.
/
/ This library is distributed in the hope that it will be useful,
/ but WITHOUT ANY WARRANTY; without even the implied warranty of
/ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/ Lesser General Public License for more details.
/
/ You should have received a copy of the GNU Lesser General Public
/ License along with this library; if not, write to the Free Software
/ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA/
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include "sqlite3ext.h"
#include "spatialite.h"

SPATIALITE_DECLARE char *
sple_clean_wkt (const unsigned char *old)
{
/* cleans and normalizes the WKT encoded string before effective parsing */
    int len;
    char *buf;
    char *pn;
    int error = 0;
    int space = 1;
    int opened = 0;
    int closed = 0;
    int ok;
    const unsigned char *po;
    len = strlen ((char *) old);
    if (len == 0)
	return NULL;
    buf = malloc (len + 1);
    pn = buf;
    po = old;
    while (*po != '\0')
      {
	  if (*po >= '0' && *po <= '9')
	    {
		*pn = *po;
		pn++;
		space = 0;
	    }
	  else if ((*po >= 'A' && *po <= 'Z') ||
		   (*po >= 'a' && *po <= 'z') ||
		   *po == ',' || *po == '.' || *po == '(' || *po == ')')
	    {
		if (pn > buf)
		  {
		      if (*(pn - 1) == ' ')
			  *(pn - 1) = *po;
		      else
			{
			    *pn = *po;
			    pn++;
			}
		  }
		else
		  {
		      *pn = *po;
		      pn++;
		  }
		if (*po == '(')
		    opened++;
		if (*po == ')')
		    closed++;
		space = 1;
	    }
	  else if (*po == ' ' || *po == '\t' || *po == '\n' || *po == '\r')
	    {
		if (!space)
		  {
		      *pn = ' ';
		      pn++;
		  }
		space = 1;
	    }
	  else
	    {
		error = 1;
		break;
	    }
	  po++;
      }
    if (opened != closed)
	error = 1;
    *pn = '\0';
    len = strlen (buf);
    if (buf[len - 1] != ')')
	error = 1;
    ok = 0;
    if (!error)
      {
	  if (len > 6 && strncasecmp (buf, "POINT(", 6) == 0)
	      ok = 1;
	  if (len > 11 && strncasecmp (buf, "LINESTRING(", 11) == 0)
	      ok = 1;
	  if (len > 8 && strncasecmp (buf, "POLYGON(", 8) == 0)
	      ok = 1;
	  if (len > 11 && strncasecmp (buf, "MULTIPOINT(", 11) == 0)
	      ok = 1;
	  if (len > 16 && strncasecmp (buf, "MULTILINESTRING(", 16) == 0)
	      ok = 1;
	  if (len > 13 && strncasecmp (buf, "MULTIPOLYGON(", 13) == 0)
	      ok = 1;
	  if (len > 19 && strncasecmp (buf, "GEOMETRYCOLLECTION(", 19) == 0)
	      ok = 1;
	  if (!ok)
	      error = 1;
      }
    if (error)
      {
	  free (buf);
	  return NULL;
      }
    return buf;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_parse_wkt (char *buffer, short type)
{
/* tryes to build a GEOMETRY by parsing a WKT encoded string */
    sple_token_ptr first = NULL;
    sple_token_ptr last = NULL;
    sple_token_ptr pt;
    sple_token_ptr ptn;
    char dummy[256];
    char *po = dummy;
    char *p = buffer;
    int opened;
    int max_opened;
    int closed;
    sple_list_token_ptr list_token = NULL;
    sple_multi_list_token_ptr multi_list_token = NULL;
    sple_multi_multi_list_token_ptr multi_multi_list_token = NULL;
    sple_geocoll_list_token_ptr geocoll_list_token = NULL;
    sple_point_ptr point;
    sple_linestring_ptr line;
    sple_geometry_ptr geo = NULL;
    while (*p != '\0')
      {
	  /* break the WKT string into tokens */
	  if (*p == '(')
	    {
		*po = '\0';
		sple_add_token (dummy, &first, &last);
		sple_add_token ("(", &first, &last);
		po = dummy;
		p++;
		continue;
	    }
	  if (*p == ')')
	    {
		*po = '\0';
		sple_add_token (dummy, &first, &last);
		sple_add_token (")", &first, &last);
		po = dummy;
		p++;
		continue;
	    }
	  if (*p == ' ')
	    {
		*po = '\0';
		sple_add_token (dummy, &first, &last);
		sple_add_token (" ", &first, &last);
		po = dummy;
		p++;
		continue;
	    }
	  if (*p == ',')
	    {
		*po = '\0';
		sple_add_token (dummy, &first, &last);
		sple_add_token (",", &first, &last);
		po = dummy;
		p++;
		continue;
	    }
	  *po++ = *p++;
      }
    if (first == NULL)
	goto err;
    max_opened = 0;
    opened = 0;
    closed = 0;
    pt = first;
    while (pt != NULL)
      {
	  /* checks for absence of serious errors */
	  if (pt->type == SPLE_UNKNOWN)
	      goto err;
	  if (pt->type == SPLE_OPENED)
	    {
		opened++;
		if (opened > 3)
		    goto err;
		if (opened > max_opened)
		    max_opened = opened;
	    }
	  if (pt->type == SPLE_CLOSED)
	    {
		closed++;
		if (closed > opened)
		    goto err;
		opened--;
		closed--;
	    }
	  pt = pt->next;
      }
    if (opened == 0 && closed == 0)
	;
    else
	goto err;
    if (type < 0)
	;			/* no restrinction about GEOMETRY CLASS TYPE */
    else
      {
	  if (first->type != type)
	      goto err;		/* invalid CLASS TYPE for request */
      }
/* preparing the organized token-list structures */
    switch (first->type)
      {
      case SPLE_POINT:
	  if (max_opened != 1)
	      goto err;
	  break;
      case SPLE_LINESTRING:
	  if (max_opened != 1)
	      goto err;
	  list_token = sple_build_list_token (first->next, last);
	  break;
      case SPLE_POLYGON:
	  if (max_opened != 2)
	      goto err;
	  multi_list_token = sple_build_multi_list_token (first->next, last);
	  break;
      case SPLE_MULTIPOINT:
	  if (max_opened != 1)
	      goto err;
	  list_token = sple_build_list_token (first->next, last);
	  break;
      case SPLE_MULTILINESTRING:
	  if (max_opened != 2)
	      goto err;
	  multi_list_token = sple_build_multi_list_token (first->next, last);
	  break;
      case SPLE_MULTIPOLYGON:
	  if (max_opened != 3)
	      goto err;
	  multi_multi_list_token =
	      sple_build_multi_multi_list_token (first->next, last);
	  break;
      case SPLE_GEOMETRYCOLLECTION:
	  if (max_opened == 2 || max_opened == 3)
	      ;
	  else
	      goto err;
	  geocoll_list_token =
	      sple_build_geocoll_list_token (first->next, last);
	  break;
      }
    switch (first->type)
      {
      case SPLE_POINT:
	  point = sple_build_point (first->next);
	  if (point)
	      geo = sple_geometry_from_point (point);
	  break;
      case SPLE_LINESTRING:
	  line = list_token->line;
	  if (line)
	      geo = sple_geometry_from_linestring (line);
	  list_token->line = NULL;
	  break;
      case SPLE_POLYGON:
	  if (multi_list_token)
	      geo = sple_geometry_from_polygon (multi_list_token);
	  break;
      case SPLE_MULTIPOINT:
	  line = list_token->line;
	  if (line)
	      geo = sple_geometry_from_mpoint (line);
	  list_token->line = NULL;
	  break;
      case SPLE_MULTILINESTRING:
	  if (multi_list_token)
	      geo = sple_geometry_from_mline (multi_list_token);
	  break;
      case SPLE_MULTIPOLYGON:
	  if (multi_multi_list_token)
	      geo = sple_geometry_from_mpoly (multi_multi_list_token);
	  break;
      case SPLE_GEOMETRYCOLLECTION:
	  if (geocoll_list_token)
	      geo = sple_geometry_from_geocoll (geocoll_list_token);
	  break;
      }
    sple_free_list_token (list_token);
    sple_free_multi_list_token (multi_list_token);
    sple_free_multi_multi_list_token (multi_multi_list_token);
    sple_free_geocoll_list_token (geocoll_list_token);
    pt = first;
    while (pt != NULL)
      {
	  /* cleans the token's list */
	  ptn = pt->next;
	  free (pt);
	  pt = ptn;
      }
    return geo;
  err:
    sple_free_list_token (list_token);
    sple_free_multi_list_token (multi_list_token);
    sple_free_multi_multi_list_token (multi_multi_list_token);
    sple_free_geocoll_list_token (geocoll_list_token);
    pt = first;
    while (pt != NULL)
      {
	  /* cleans the token's list */
	  ptn = pt->next;
	  free (pt);
	  pt = ptn;
      }
    return NULL;
}

SPATIALITE_DECLARE void
sple_add_token (char *token, sple_token_ptr * first, sple_token_ptr * last)
{
/* inserts a token at the end of the linked list */
    double coord;
    sple_token_ptr p;
    if (strlen (token) == 0)
	return;
    p = malloc (sizeof (sple_token));
    p->type = SPLE_UNKNOWN;
    p->coord = 0.0;
    if (strcasecmp (token, "POINT") == 0)
	p->type = SPLE_POINT;
    if (strcasecmp (token, "LINESTRING") == 0)
	p->type = SPLE_LINESTRING;
    if (strcasecmp (token, "POLYGON") == 0)
	p->type = SPLE_POLYGON;
    if (strcasecmp (token, "MULTIPOINT") == 0)
	p->type = SPLE_MULTIPOINT;
    if (strcasecmp (token, "MULTILINESTRING") == 0)
	p->type = SPLE_MULTILINESTRING;
    if (strcasecmp (token, "MULTIPOLYGON") == 0)
	p->type = SPLE_MULTIPOLYGON;
    if (strcasecmp (token, "GEOMETRYCOLLECTION") == 0)
	p->type = SPLE_GEOMETRYCOLLECTION;
    if (strcmp (token, "(") == 0)
	p->type = SPLE_OPENED;
    if (strcmp (token, ")") == 0)
	p->type = SPLE_CLOSED;
    if (strcmp (token, ",") == 0)
	p->type = SPLE_COMMA;
    if (strcmp (token, " ") == 0)
	p->type = SPLE_SPACE;
    if (p->type == SPLE_UNKNOWN)
      {
	  if (sple_parse_double (token, &coord))
	    {
		p->type = SPLE_COORDINATE;
		p->coord = coord;
	    }
      }
    p->next = NULL;
    if (*first == NULL)
	*first = p;
    if (*last != NULL)
	(*last)->next = p;
    *last = p;
}

SPATIALITE_DECLARE int
sple_parse_double (char *token, double *coord)
{
/* checks if this token is a valid double */
    int i;
    int digits = 0;
    int errs = 0;
    int commas = 0;
    *coord = 0.0;
    for (i = 0; i < strlen (token); i++)
      {
	  if (token[i] == '.')
	      commas++;
	  else if (token[i] >= '0' && token[i] <= '9')
	      digits++;
	  else
	      errs++;
      }
    if (errs > 0)
	return 0;
    if (digits == 0)
	return 0;
    if (commas > 1)
	return 0;
    *coord = atof (token);
    return 1;
}

SPATIALITE_DECLARE sple_point_ptr
sple_build_point (sple_token_ptr first)
{
/* builds a POINT, if this token's list contains a valid POINT */
    sple_point_ptr point = NULL;
    sple_token_ptr pt = first;
    int i = 0;
    int err = 0;
    double x = 0.0;
    double y = 0.0;
    while (pt != NULL)
      {
	  /* check if this one is a valid POINT */
	  switch (i)
	    {
	    case 0:
		if (pt->type != SPLE_OPENED)
		    err = 1;
		break;
	    case 1:
		if (pt->type != SPLE_COORDINATE)
		    err = 1;
		else
		    x = pt->coord;
		break;
	    case 2:
		if (pt->type != SPLE_SPACE)
		    err = 1;
		break;
	    case 3:
		if (pt->type != SPLE_COORDINATE)
		    err = 1;
		else
		    y = pt->coord;
		break;
	    case 4:
		if (pt->type != SPLE_CLOSED)
		    err = 1;
		break;
	    default:
		err = 1;
		break;
	    };
	  i++;
	  pt = pt->next;
      }
    if (err)
	return NULL;
    point = sple_alloc_point (x, y);
    return point;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_geometry_from_point (sple_point_ptr point)
{
/* builds a GEOMETRY containing a POINT */
    sple_geometry_ptr geom = NULL;
    geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = 1;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(geom->geometries) =
	sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
    return geom;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_geometry_from_mpoint (sple_linestring_ptr mpoint)
{
/* builds a GEOMETRY containing a MULTIPOINT */
    int ie;
    sple_point_ptr point;
    sple_geometry_ptr geom = NULL;
    geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = mpoint->points;
    geom->geometries =
	malloc (sizeof (sple_generic_geometry_ptr) * mpoint->points);
    for (ie = 0; ie < mpoint->points; ie++)
      {
	  point =
	      sple_alloc_point ((mpoint->geometry + ie)->x,
				(mpoint->geometry + ie)->y);
	  *(geom->geometries + ie) =
	      sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
      }
    return geom;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_geometry_from_linestring (sple_linestring_ptr line)
{
/* builds a GEOMETRY containing a LINESTRING */
    sple_geometry_ptr geom = NULL;
    geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = 1;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(geom->geometries) =
	sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) line);
    return geom;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_geometry_from_mline (sple_multi_list_token_ptr mline)
{
/* builds a GEOMETRY containing a MULTILINESTRING */
    int ie;
    int iv;
    int lines = 0;
    sple_list_token_ptr pt;
    sple_linestring_ptr line;
    sple_linestring_ptr line2;
    sple_geometry_ptr geom = NULL;
    pt = mline->first;
    while (pt != NULL)
      {
/* counts how many linestrings are in the list */
	  lines++;
	  pt = pt->next;
      }
    if (!lines)
	return NULL;
    geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = lines;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr) * lines);
    ie = 0;
    pt = mline->first;
    while (pt != NULL)
      {
	  /* creates and initializes one linestring for each iteration */
	  line = pt->line;
	  line2 = sple_alloc_linestring (line->points);
	  for (iv = 0; iv < line->points; iv++)
	      sple_set_point (line2, iv, (line->geometry + iv)->x,
			      (line->geometry + iv)->y);
	  *(geom->geometries + ie) =
	      sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) line2);
	  ie++;
	  pt = pt->next;
      }
    return geom;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_geometry_from_polygon (sple_multi_list_token_ptr polygon)
{
/* builds a GEOMETRY containing a POLYGON */
    int iv;
    int ib;
    int borders = 0;
    sple_polygon_ptr pg;
    sple_linestring_ptr ring;
    sple_linestring_ptr line;
    sple_geometry_ptr geom = NULL;
    sple_list_token_ptr pt;
    pt = polygon->first;
    while (pt != NULL)
      {
	  /* counts how many rings are in the list */
	  borders++;
	  pt = pt->next;
      }
    if (!borders)
	return NULL;
/* builds the polygon */
    ring = polygon->first->line;
    pg = sple_alloc_polygon (ring->points, borders - 1);
    line = pg->external;
    for (iv = 0; iv < line->points; iv++)
      {
	  /* sets the POINTS for the exterior ring */
	  sple_set_point (line, iv, (ring->geometry + iv)->x,
			  (ring->geometry + iv)->y);
      }
    ib = 0;
    pt = polygon->first->next;
    while (pt != NULL)
      {
	  /* builds the interior rings [if any] */
	  ring = pt->line;
	  line = sple_alloc_internal_border (pg, ib, ring->points);
	  for (iv = 0; iv < line->points; iv++)
	      sple_set_point (line, iv, (ring->geometry + iv)->x,
			      (ring->geometry + iv)->y);
	  ib++;
	  pt = pt->next;
      }
/* allocates and initializes the geometry to be returned */
    geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = 1;
/* inserts the polygon inside the geometry to be returned */
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
    *(geom->geometries) =
	sple_alloc_generic_geometry (SPLE_POLYGON, (void *) pg);
    return geom;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_geometry_from_mpoly (sple_multi_multi_list_token_ptr mpoly)
{
/* builds a GEOMETRY containing a MULTIPOLYGON */
    int iv;
    int ib;
    int ie;
    int borders;
    int entities = 0;
    sple_polygon_ptr pg;
    sple_linestring_ptr ring;
    sple_linestring_ptr line;
    sple_geometry_ptr geom = NULL;
    sple_multi_list_token_ptr multi;
    sple_list_token_ptr pt;
    multi = mpoly->first;
    while (multi != NULL)
      {
	  /* counts how many polygons are in the list */
	  entities++;
	  multi = multi->next;
      }
    if (!entities)
	return NULL;
/* allocates and initializes the geometry to be returned */
    geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = entities;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr) * entities);
    ie = 0;
    multi = mpoly->first;
    while (multi != NULL)
      {
	  borders = 0;
	  pt = multi->first;
	  while (pt != NULL)
	    {
		/* counts how many rings are in the list */
		borders++;
		pt = pt->next;
	    }
	  /* builds one polygon */
	  ring = multi->first->line;
	  pg = sple_alloc_polygon (ring->points, borders - 1);
	  line = pg->external;
	  for (iv = 0; iv < line->points; iv++)
	    {
		/* sets the POINTS for the exterior ring */
		sple_set_point (line, iv, (ring->geometry + iv)->x,
				(ring->geometry + iv)->y);
	    }
	  ib = 0;
	  pt = multi->first->next;
	  while (pt != NULL)
	    {
		/* builds the interior rings [if any] */
		ring = pt->line;
		line = sple_alloc_internal_border (pg, ib, ring->points);
		for (iv = 0; iv < line->points; iv++)
		    sple_set_point (line, iv, (ring->geometry + iv)->x,
				    (ring->geometry + iv)->y);
		ib++;
		pt = pt->next;
	    }
	  /* inserts the polygon inside the geometry to be returned */
	  *(geom->geometries + ie) =
	      sple_alloc_generic_geometry (SPLE_POLYGON, (void *) pg);
	  ie++;
	  multi = multi->next;
      }
    return geom;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_geometry_from_geocoll (sple_geocoll_list_token_ptr geocoll)
{
/* builds a GEOMETRY containing a GEOMETRYCOLLECTION */
    int iv;
    int ib;
    int ie;
    int borders;
    int entities = 0;
    sple_point_ptr point;
    sple_polygon_ptr pg;
    sple_linestring_ptr ring;
    sple_linestring_ptr line;
    sple_geometry_ptr geom = NULL;
    sple_list_token_ptr linestring;
    sple_multi_list_token_ptr polyg;
    sple_var_list_token_ptr multi;
    sple_list_token_ptr pt;
    multi = geocoll->first;
    while (multi != NULL)
      {
	  /* counts how many polygons are in the list */
	  entities++;
	  multi = multi->next;
      }
    if (!entities)
	return NULL;
/* allocates and initializes the geometry to be returned */
    geom = sple_alloc_geometry ();
    geom->little_endian = 1;
    geom->blob = NULL;
    geom->size = 0;
    geom->offset = 0;
    geom->entities = entities;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr) * entities);
    ie = 0;
    multi = geocoll->first;
    while (multi != NULL)
      {
	  switch (multi->type)
	    {
	    case SPLE_POINT:
		point = sple_alloc_point (multi->x, multi->y);
		*(geom->geometries + ie) =
		    sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
		break;
	    case SPLE_LINESTRING:
		linestring = (sple_list_token_ptr) (multi->pointer);
		ring = linestring->line;
		line = sple_alloc_linestring (ring->points);
		for (iv = 0; iv < line->points; iv++)
		  {
		      /* sets the POINTS for the LINESTRING */
		      sple_set_point (line, iv, (ring->geometry + iv)->x,
				      (ring->geometry + iv)->y);
		  }
		*(geom->geometries + ie) =
		    sple_alloc_generic_geometry (SPLE_LINESTRING,
						 (void *) line);
		break;
	    case SPLE_POLYGON:
		polyg = multi->pointer;
		borders = 0;
		pt = polyg->first;
		while (pt != NULL)
		  {
		      /* counts how many rings are in the list */
		      borders++;
		      pt = pt->next;
		  }
		/* builds one polygon */
		ring = polyg->first->line;
		pg = sple_alloc_polygon (ring->points, borders - 1);
		line = pg->external;
		for (iv = 0; iv < line->points; iv++)
		  {
		      /* sets the POINTS for the exterior ring */
		      sple_set_point (line, iv, (ring->geometry + iv)->x,
				      (ring->geometry + iv)->y);
		  }
		ib = 0;
		pt = polyg->first->next;
		while (pt != NULL)
		  {
		      /* builds the interior rings [if any] */
		      ring = pt->line;
		      line = sple_alloc_internal_border (pg, ib, ring->points);
		      for (iv = 0; iv < line->points; iv++)
			  sple_set_point (line, iv, (ring->geometry + iv)->x,
					  (ring->geometry + iv)->y);
		      ib++;
		      pt = pt->next;
		  }
		/* inserts the polygon inside the geometry to be returned */
		*(geom->geometries + ie) =
		    sple_alloc_generic_geometry (SPLE_POLYGON, (void *) pg);
		break;
	    };
	  ie++;
	  multi = multi->next;
      }
    return geom;
}

SPATIALITE_DECLARE sple_list_token_ptr
sple_build_list_token (sple_token_ptr first, sple_token_ptr last)
{
/* builds a list of tokens representing a list in the form (1 2, 3 4, 5 6), as required by LINESTRING, MULTIPOINT or RING */
    sple_list_token_ptr list = NULL;
    sple_token_ptr pt;
    int i = 0;
    int ip = 0;
    int err = 0;
    int nx = 0;
    int ny = 0;
    int iv;
    double x = 0.0;
    double y;
    pt = first;
    pt = first;
    while (pt != NULL)
      {
	  /* check if this one is a valid list of POINTS */
	  if (i == 0)
	    {
		if (pt->type != SPLE_OPENED)
		    err = 1;
	    }
	  else if (pt == last)
	    {
		if (pt->type != SPLE_CLOSED)
		    err = 1;
	    }
	  else
	    {
		if (ip == 0)
		  {
		      if (pt->type != SPLE_COORDINATE)
			  err = 1;
		      else
			  nx++;
		  }
		else if (ip == 1)
		  {
		      if (pt->type != SPLE_SPACE)
			  err = 1;
		  }
		else if (ip == 2)
		  {
		      if (pt->type != SPLE_COORDINATE)
			  err = 1;
		      else
			  ny++;
		  }
		else if (ip == 3)
		  {
		      if (pt->type != SPLE_COMMA)
			  err = 1;
		  }
		ip++;
		if (ip > 3)
		    ip = 0;
	    }
	  i++;
	  pt = pt->next;
	  if (pt == last)
	      break;
      }
    if (nx != ny)
	err = 1;
    if (nx < 2)
	err = 1;
    if (err)
	return NULL;
/* ok, there is no erorr. finally we can build the POINTS list */
    list = malloc (sizeof (sple_list_token));
    list->points = nx;
    list->line = sple_alloc_linestring (nx);
    list->next = NULL;
    iv = 0;
    ip = 0;
    i = 0;
    pt = first;
    while (pt != NULL)
      {
	  /* sets coords for all POINTS */
	  if (i == 0)
	      ;
	  else if (pt == last)
	      ;
	  else
	    {
		if (ip == 0)
		    x = pt->coord;
		else if (ip == 2)
		  {
		      y = pt->coord;
		      sple_set_point (list->line, iv++, x, y);
		  }
		ip++;
		if (ip > 3)
		    ip = 0;
	    }
	  i++;
	  pt = pt->next;
	  if (pt == last)
	      break;
      }
    return list;
}

SPATIALITE_DECLARE void
sple_free_list_token (sple_list_token_ptr p)
{
/* cleans all memory allocations for list token */
    if (!p)
	return;
    if (p->line)
	sple_free_linestring (p->line);
    free (p);
}

SPATIALITE_DECLARE sple_multi_list_token_ptr
sple_build_multi_list_token (sple_token_ptr first, sple_token_ptr last)
{
/* builds a multi list of tokens representing an array of elementar lists in the form ((...),(....),(...)), as required by MULTILINESTRING and POLYGON */
    sple_multi_list_token_ptr multi_list = NULL;
    sple_token_ptr pt;
    sple_token_ptr p_first = NULL;
    sple_list_token_ptr list;
    int opened = 0;
    pt = first;
    while (pt != NULL)
      {
	  /* identifies the sub-lists contained in this multi list */
	  if (pt->type == SPLE_OPENED)
	    {
		opened++;
		if (opened == 2)
		    p_first = pt;
	    }
	  if (pt->type == SPLE_CLOSED)
	    {
		if (p_first)
		  {
		      list = sple_build_list_token (p_first, pt);
		      if (!multi_list)
			{
			    multi_list =
				malloc (sizeof (sple_multi_list_token));
			    multi_list->first = NULL;
			    multi_list->last = NULL;
			    multi_list->next = NULL;
			}
		      if (multi_list->first == NULL)
			  multi_list->first = list;
		      if (multi_list->last != NULL)
			  multi_list->last->next = list;
		      multi_list->last = list;
		      p_first = NULL;
		  }
		opened--;
	    }
	  pt = pt->next;
	  if (pt == last)
	      break;
      }
    return multi_list;
}

SPATIALITE_DECLARE void
sple_free_multi_list_token (sple_multi_list_token_ptr p)
{
/* cleans all memory allocations for multi list token */
    sple_list_token_ptr pt;
    sple_list_token_ptr ptn;
    if (!p)
	return;
    pt = p->first;
    while (pt)
      {
	  ptn = pt->next;
	  sple_free_list_token (pt);
	  pt = ptn;
      }
    free (p);
}

SPATIALITE_DECLARE sple_multi_multi_list_token_ptr
sple_build_multi_multi_list_token (sple_token_ptr first, sple_token_ptr last)
{
/* builds a multi list of tokens representing an array of complex lists in the form (((...),(....),(...)),((...),(...)),((...))), as required by MULTIPOLYGON */
    sple_multi_multi_list_token_ptr multi_multi_list = NULL;
    sple_token_ptr pt;
    sple_token_ptr p_first = NULL;
    sple_multi_list_token_ptr multi_list;
    int opened = 0;
    pt = first;
    while (pt != NULL)
      {
	  /* identifies the sub-lists contained in this multi list */
	  if (pt->type == SPLE_OPENED)
	    {
		opened++;
		if (opened == 2)
		    p_first = pt;
	    }
	  if (pt->type == SPLE_CLOSED)
	    {
		if (p_first && opened == 2)
		  {
		      multi_list = sple_build_multi_list_token (p_first, pt);
		      if (!multi_multi_list)
			{
			    multi_multi_list =
				malloc (sizeof (sple_multi_multi_list_token));
			    multi_multi_list->first = NULL;
			    multi_multi_list->last = NULL;
			}
		      if (multi_multi_list->first == NULL)
			  multi_multi_list->first = multi_list;
		      if (multi_multi_list->last != NULL)
			  multi_multi_list->last->next = multi_list;
		      multi_multi_list->last = multi_list;
		      p_first = NULL;
		  }
		opened--;
	    }
	  pt = pt->next;
	  if (pt == last)
	      break;
      }
    return multi_multi_list;
}

SPATIALITE_DECLARE void
sple_free_multi_multi_list_token (sple_multi_multi_list_token_ptr p)
{
/* cleans all memory allocations for multi-multi list token */
    sple_multi_list_token_ptr pt;
    sple_multi_list_token_ptr ptn;
    if (!p)
	return;
    pt = p->first;
    while (pt)
      {
	  ptn = pt->next;
	  sple_free_multi_list_token (pt);
	  pt = ptn;
      }
    free (p);
}

SPATIALITE_DECLARE sple_geocoll_list_token_ptr
sple_build_geocoll_list_token (sple_token_ptr first, sple_token_ptr last)
{
/* builds a variable list of tokens representing an array of entities in the form (ELEM(),ELEM(),ELEM())  as required by GEOMETRYCOLLECTION */
    sple_geocoll_list_token_ptr geocoll_list = NULL;
    sple_token_ptr pt;
    sple_token_ptr pt2;
    sple_token_ptr p_first = NULL;
    sple_list_token_ptr list;
    sple_multi_list_token_ptr multi_list;
    sple_var_list_token_ptr var_list;
    int opened = 0;
    int i;
    int err;
    double x = 0;
    double y = 0;
    pt = first;
    while (pt != NULL)
      {
	  /* identifies the sub-lists contained in this complex list */
	  if (pt->type == SPLE_POINT)
	    {
		/* parsing a POINT list */
		err = 0;
		i = 0;
		pt2 = pt->next;
		while (pt2 != NULL)
		  {
		      /* check if this one is a valid POINT */
		      switch (i)
			{
			case 0:
			    if (pt2->type != SPLE_OPENED)
				err = 1;
			    break;
			case 1:
			    if (pt2->type != SPLE_COORDINATE)
				err = 1;
			    else
				x = pt2->coord;
			    break;
			case 2:
			    if (pt2->type != SPLE_SPACE)
				err = 1;
			    break;
			case 3:
			    if (pt2->type != SPLE_COORDINATE)
				err = 1;
			    else
				y = pt2->coord;
			    break;
			case 4:
			    if (pt2->type != SPLE_CLOSED)
				err = 1;
			    break;
			};
		      i++;
		      if (i > 4)
			  break;
		      pt2 = pt2->next;
		  }
		if (err)
		    goto error;
		var_list = malloc (sizeof (sple_var_list_token));
		var_list->type = SPLE_POINT;
		var_list->x = x;
		var_list->y = y;
		var_list->next = NULL;
		if (!geocoll_list)
		  {
		      geocoll_list = malloc (sizeof (sple_geocoll_list_token));
		      geocoll_list->first = NULL;
		      geocoll_list->last = NULL;
		  }
		if (geocoll_list->first == NULL)
		    geocoll_list->first = var_list;
		if (geocoll_list->last != NULL)
		    geocoll_list->last->next = var_list;
		geocoll_list->last = var_list;
	    }
	  else if (pt->type == SPLE_LINESTRING)
	    {
		/* parsing a LINESTRING list */
		p_first = NULL;
		pt2 = pt->next;
		while (pt2 != NULL)
		  {
		      if (pt2->type == SPLE_OPENED)
			  p_first = pt2;
		      if (pt2->type == SPLE_CLOSED)
			{
			    list = sple_build_list_token (p_first, pt2);
			    if (list)
			      {
				  var_list =
				      malloc (sizeof (sple_var_list_token));
				  var_list->type = SPLE_LINESTRING;
				  var_list->pointer = list;
				  var_list->next = NULL;
				  if (!geocoll_list)
				    {
					geocoll_list =
					    malloc (sizeof
						    (sple_geocoll_list_token));
					geocoll_list->first = NULL;
					geocoll_list->last = NULL;
				    }
				  if (geocoll_list->first == NULL)
				      geocoll_list->first = var_list;
				  if (geocoll_list->last != NULL)
				      geocoll_list->last->next = var_list;
				  geocoll_list->last = var_list;
				  break;
			      }
			    else
				goto error;
			}
		      pt2 = pt2->next;
		      if (pt2 == last)
			  break;
		  }
	    }
	  else if (pt->type == SPLE_POLYGON)
	    {
		/* parsing a POLYGON list */
		opened = 0;
		p_first = NULL;
		pt2 = pt->next;
		while (pt2 != NULL)
		  {
		      if (pt2->type == SPLE_OPENED)
			{
			    opened++;
			    if (opened == 1)
				p_first = pt2;
			}
		      if (pt2->type == SPLE_CLOSED)
			{
			    if (p_first && opened == 1)
			      {
				  multi_list =
				      sple_build_multi_list_token (p_first,
								   pt2);
				  if (multi_list)
				    {
					var_list =
					    malloc (sizeof
						    (sple_var_list_token));
					var_list->type = SPLE_POLYGON;
					var_list->pointer = multi_list;
					var_list->next = NULL;
					if (!geocoll_list)
					  {
					      geocoll_list =
						  malloc (sizeof
							  (sple_geocoll_list_token));
					      geocoll_list->first = NULL;
					      geocoll_list->last = NULL;
					  }
					if (geocoll_list->first == NULL)
					    geocoll_list->first = var_list;
					if (geocoll_list->last != NULL)
					    geocoll_list->last->next = var_list;
					geocoll_list->last = var_list;
					break;
				    }
				  else
				      goto error;
			      }
			    opened--;
			}
		      pt2 = pt2->next;
		      if (pt2 == last)
			  break;
		  }
	    }
	  pt = pt->next;
      }
    return geocoll_list;
  error:
    sple_free_geocoll_list_token (geocoll_list);
    return NULL;
}

SPATIALITE_DECLARE void
sple_free_geocoll_list_token (sple_geocoll_list_token_ptr p)
{
/* cleans all memory allocations for geocoll list token */
    sple_var_list_token_ptr pt;
    sple_var_list_token_ptr ptn;
    if (!p)
	return;
    pt = p->first;
    while (pt)
      {
	  ptn = pt->next;
	  if (pt->type == SPLE_LINESTRING)
	      sple_free_multi_list_token ((sple_multi_list_token_ptr)
					  (pt->pointer));
	  if (pt->type == SPLE_POLYGON)
	      sple_free_multi_multi_list_token ((sple_multi_multi_list_token_ptr) (pt->pointer));
	  pt = ptn;
      }
    free (p);
}
